import numpy as np
import pandas as pd
import pytest

from featuretools.primitives import (
    MaxConsecutiveFalse,
    MaxConsecutiveNegatives,
    MaxConsecutivePositives,
    MaxConsecutiveTrue,
    MaxConsecutiveZeros,
)


class TestMaxConsecutiveFalse:
    def test_regular(self):
        primitive_instance = MaxConsecutiveFalse()
        primitive_func = primitive_instance.get_function()
        array = pd.Series([False, False, False, True, True, False, True], dtype="bool")
        assert primitive_func(array) == 3

    def test_all_true(self):
        primitive_instance = MaxConsecutiveFalse()
        primitive_func = primitive_instance.get_function()
        array = pd.Series([True, True, True, True], dtype="bool")
        assert primitive_func(array) == 0

    def test_all_false(self):
        primitive_instance = MaxConsecutiveFalse()
        primitive_func = primitive_instance.get_function()
        array = pd.Series([False, False, False], dtype="bool")
        assert primitive_func(array) == 3


class TestMaxConsecutiveTrue:
    def test_regular(self):
        primitive_instance = MaxConsecutiveTrue()
        primitive_func = primitive_instance.get_function()
        array = pd.Series([True, False, True, True, True, False, True], dtype="bool")
        assert primitive_func(array) == 3

    def test_all_true(self):
        primitive_instance = MaxConsecutiveTrue()
        primitive_func = primitive_instance.get_function()
        array = pd.Series([True, True, True, True], dtype="bool")
        assert primitive_func(array) == 4

    def test_all_false(self):
        primitive_instance = MaxConsecutiveTrue()
        primitive_func = primitive_instance.get_function()
        array = pd.Series([False, False, False], dtype="bool")
        assert primitive_func(array) == 0


@pytest.mark.parametrize("dtype", ["float64", "int64"])
class TestMaxConsecutiveNegatives:
    def test_regular(self, dtype):
        if dtype == "int64":
            pytest.skip("test array contains floats which are not supported int64")
        primitive_instance = MaxConsecutiveNegatives()
        primitive_func = primitive_instance.get_function()
        array = pd.Series([1.3, -3.4, -1, -4, 10, -1.7, -4.9], dtype=dtype)
        assert primitive_func(array) == 3

    def test_all_int(self, dtype):
        primitive_instance = MaxConsecutiveNegatives()
        primitive_func = primitive_instance.get_function()
        array = pd.Series([1, -1, 2, 4, -5], dtype=dtype)
        assert primitive_func(array) == 1

    def test_all_float(self, dtype):
        if dtype == "int64":
            pytest.skip("test array contains floats which are not supported int64")
        primitive_instance = MaxConsecutiveNegatives()
        primitive_func = primitive_instance.get_function()
        array = pd.Series([1.0, -1.0, -2.0, 0.0, 5.0], dtype=dtype)
        assert primitive_func(array) == 2

    def test_with_nan(self, dtype):
        if dtype == "int64":
            pytest.skip("nans not supported in int64")
        primitive_instance = MaxConsecutiveNegatives()
        primitive_func = primitive_instance.get_function()
        array = pd.Series([1, np.nan, -2, -3], dtype=dtype)
        assert primitive_func(array) == 2

    def test_with_nan_skipna(self, dtype):
        if dtype == "int64":
            pytest.skip("nans not supported in int64")
        primitive_instance = MaxConsecutiveNegatives(skipna=False)
        primitive_func = primitive_instance.get_function()
        array = pd.Series([-1, np.nan, -2, -3], dtype=dtype)
        assert primitive_func(array) == 2

    def test_all_nan(self, dtype):
        if dtype == "int64":
            pytest.skip("nans not supported in int64")
        primitive_instance = MaxConsecutiveNegatives()
        primitive_func = primitive_instance.get_function()
        array = pd.Series([np.nan, np.nan, np.nan, np.nan], dtype=dtype)
        assert np.isnan(primitive_func(array))

    def test_all_nan_skipna(self, dtype):
        if dtype == "int64":
            pytest.skip("nans not supported in int64")
        primitive_instance = MaxConsecutiveNegatives(skipna=True)
        primitive_func = primitive_instance.get_function()
        array = pd.Series([np.nan, np.nan, np.nan, np.nan], dtype=dtype)
        assert np.isnan(primitive_func(array))


@pytest.mark.parametrize("dtype", ["float64", "int64"])
class TestMaxConsecutivePositives:
    def test_regular(self, dtype):
        if dtype == "int64":
            pytest.skip("test array contains floats which are not supported int64")
        primitive_instance = MaxConsecutivePositives()
        primitive_func = primitive_instance.get_function()
        array = pd.Series([1.3, -3.4, 1, 4, 10, -1.7, -4.9], dtype=dtype)
        assert primitive_func(array) == 3

    def test_all_int(self, dtype):
        primitive_instance = MaxConsecutivePositives()
        primitive_func = primitive_instance.get_function()
        array = pd.Series([1, -1, 2, 4, -5], dtype=dtype)
        assert primitive_func(array) == 2

    def test_all_float(self, dtype):
        if dtype == "int64":
            pytest.skip("test array contains floats which are not supported int64")
        primitive_instance = MaxConsecutivePositives()
        primitive_func = primitive_instance.get_function()
        array = pd.Series([1.0, -1.0, 2.0, 4.0, 5.0], dtype=dtype)
        assert primitive_func(array) == 3

    def test_with_nan(self, dtype):
        if dtype == "int64":
            pytest.skip("nans not supported in int64")
        primitive_instance = MaxConsecutivePositives()
        primitive_func = primitive_instance.get_function()
        array = pd.Series([1, np.nan, 2, -3], dtype=dtype)
        assert primitive_func(array) == 2

    def test_with_nan_skipna(self, dtype):
        if dtype == "int64":
            pytest.skip("nans not supported in int64")
        primitive_instance = MaxConsecutivePositives(skipna=False)
        primitive_func = primitive_instance.get_function()
        array = pd.Series([1, np.nan, 2, -3], dtype=dtype)
        assert primitive_func(array) == 1

    def test_all_nan(self, dtype):
        if dtype == "int64":
            pytest.skip("nans not supported in int64")
        primitive_instance = MaxConsecutivePositives()
        primitive_func = primitive_instance.get_function()
        array = pd.Series([np.nan, np.nan, np.nan, np.nan], dtype=dtype)
        assert np.isnan(primitive_func(array))

    def test_all_nan_skipna(self, dtype):
        if dtype == "int64":
            pytest.skip("nans not supported in int64")
        primitive_instance = MaxConsecutivePositives(skipna=True)
        primitive_func = primitive_instance.get_function()
        array = pd.Series([np.nan, np.nan, np.nan, np.nan], dtype=dtype)
        assert np.isnan(primitive_func(array))


@pytest.mark.parametrize("dtype", ["float64", "int64"])
class TestMaxConsecutiveZeros:
    def test_regular(self, dtype):
        if dtype == "int64":
            pytest.skip("test array contains floats which are not supported int64")
        primitive_instance = MaxConsecutiveZeros()
        primitive_func = primitive_instance.get_function()
        array = pd.Series([1.3, -3.4, 0, 0, 0.0, 1.7, -4.9], dtype=dtype)
        assert primitive_func(array) == 3

    def test_all_int(self, dtype):
        primitive_instance = MaxConsecutiveZeros()
        primitive_func = primitive_instance.get_function()
        array = pd.Series([1, -1, 0, 0, -5], dtype=dtype)
        assert primitive_func(array) == 2

    def test_all_float(self, dtype):
        if dtype == "int64":
            pytest.skip("test array contains floats which are not supported int64")
        primitive_instance = MaxConsecutiveZeros()
        primitive_func = primitive_instance.get_function()
        array = pd.Series([1.0, 0.0, 0.0, 0.0, -5.3], dtype=dtype)
        assert primitive_func(array) == 3

    def test_with_nan(self, dtype):
        if dtype == "int64":
            pytest.skip("nans not supported in int64")
        primitive_instance = MaxConsecutiveZeros()
        primitive_func = primitive_instance.get_function()
        array = pd.Series([0, np.nan, 0, -3], dtype=dtype)
        assert primitive_func(array) == 2

    def test_with_nan_skipna(self, dtype):
        if dtype == "int64":
            pytest.skip("nans not supported in int64")
        primitive_instance = MaxConsecutiveZeros(skipna=False)
        primitive_func = primitive_instance.get_function()
        array = pd.Series([0, np.nan, 0, -3], dtype=dtype)
        assert primitive_func(array) == 1

    def test_all_nan(self, dtype):
        if dtype == "int64":
            pytest.skip("nans not supported in int64")
        primitive_instance = MaxConsecutiveZeros()
        primitive_func = primitive_instance.get_function()
        array = pd.Series([np.nan, np.nan, np.nan, np.nan], dtype=dtype)
        assert np.isnan(primitive_func(array))

    def test_all_nan_skipna(self, dtype):
        if dtype == "int64":
            pytest.skip("nans not supported in int64")
        primitive_instance = MaxConsecutiveZeros(skipna=True)
        primitive_func = primitive_instance.get_function()
        array = pd.Series([np.nan, np.nan, np.nan, np.nan], dtype=dtype)
        assert np.isnan(primitive_func(array))



from woodwork.column_schema import ColumnSchema
from woodwork.logical_types import BooleanNullable, Datetime

from featuretools.primitives.base import TransformPrimitive
from featuretools.utils.gen_utils import Library


class IsWeekend(TransformPrimitive):
    """Determines if a date falls on a weekend.

    Examples:
        >>> from datetime import datetime
        >>> dates = [datetime(2019, 3, 1),
        ...          datetime(2019, 6, 17, 11, 10, 50),
        ...          datetime(2019, 11, 30, 19, 45, 15)]
        >>> is_weekend = IsWeekend()
        >>> is_weekend(dates).tolist()
        [False, False, True]
    """

    name = "is_weekend"
    input_types = [ColumnSchema(logical_type=Datetime)]
    return_type = ColumnSchema(logical_type=BooleanNullable)
    compatibility = [Library.PANDAS, Library.DASK, Library.SPARK]
    description_template = "whether {} occurred on a weekend"

    def get_function(self):
        def is_weekend(vals):
            return vals.dt.weekday > 4

        return is_weekend



import numpy as np
from woodwork.column_schema import ColumnSchema
from woodwork.logical_types import Datetime, Double

from featuretools.primitives.base import AggregationPrimitive


class FirstLastTimeDelta(AggregationPrimitive):
    """Determines the time between the first and last time value
        in seconds.

    Examples:
        >>> from datetime import datetime
        >>> first_last_time_delta = FirstLastTimeDelta()
        >>> first_last_time_delta([
        ...     datetime(2011, 4, 9, 10, 30, 0),
        ...     datetime(2011, 4, 9, 10, 30, 15),
        ...     datetime(2011, 4, 9, 10, 30, 35)])
        35.0
    """

    name = "first_last_time_delta"
    input_types = [ColumnSchema(logical_type=Datetime, semantic_tags={"time_index"})]
    return_type = ColumnSchema(logical_type=Double, semantic_tags={"numeric"})
    uses_calc_time = False
    stack_on_self = False
    default_value = 0

    def get_function(self):
        def first_last_time_delta(datetime_col):
            datetime_col = datetime_col.dropna()
            if datetime_col.empty:
                return np.nan
            delta = datetime_col.iloc[-1] - datetime_col.iloc[0]
            return delta.total_seconds()

        return first_last_time_delta



import pandas as pd
from woodwork.column_schema import ColumnSchema
from woodwork.logical_types import IntegerNullable, NaturalLanguage

from featuretools.primitives.base import TransformPrimitive
from featuretools.utils.gen_utils import Library


class NumCharacters(TransformPrimitive):
    """Calculates the number of characters in a given string, including whitespace and punctuation.

    Description:
        Returns the number of characters in a string. This is equivalent to the length of a string.

    Examples:
        >>> num_characters = NumCharacters()
        >>> num_characters(['This is a string',
        ...                 'second item',
        ...                 'final1']).tolist()
        [16, 11, 6]
    """

    name = "num_characters"
    input_types = [ColumnSchema(logical_type=NaturalLanguage)]
    return_type = ColumnSchema(logical_type=IntegerNullable, semantic_tags={"numeric"})
    compatibility = [Library.PANDAS, Library.DASK, Library.SPARK]
    description_template = "the number of characters in {}"

    def get_function(self):
        def character_counter(array):
            def _get_num_characters(elem):
                """Returns the length of elem, or pd.NA given null input"""
                if pd.isna(elem):
                    return pd.NA
                return len(elem)

            return array.apply(_get_num_characters)

        return character_counter



from woodwork.column_schema import ColumnSchema
from woodwork.logical_types import Datetime, Integer

from featuretools.primitives.base import AggregationPrimitive


class NUniqueDaysOfMonth(AggregationPrimitive):
    """Determines the number of unique days of month.

    Description:
        Given a list of datetimes, return the number of unique days
        of month. The maximum value is 31. 2018-01-01 and 2018-02-01
        will be counted as 1 unique day. 2019-01-01 and 2018-01-01
        will also be counted as 1.

    Examples:
        >>> from datetime import datetime
        >>> n_unique_days_of_month = NUniqueDaysOfMonth()
        >>> times = [datetime(2019, 1, 1),
        ...          datetime(2019, 2, 1),
        ...          datetime(2018, 2, 1),
        ...          datetime(2019, 1, 2),
        ...          datetime(2019, 1, 3)]
        >>> n_unique_days_of_month(times)
        3
    """

    name = "n_unique_days_of_month"
    input_types = [ColumnSchema(logical_type=Datetime)]
    return_type = ColumnSchema(logical_type=Integer, semantic_tags={"numeric"})
    stack_on_self = False
    default_value = 0

    def get_function(self):
        def n_unique_days_of_month(x):
            return x.dropna().dt.day.nunique()

        return n_unique_days_of_month



import numpy as np
from woodwork.column_schema import ColumnSchema

from featuretools.primitives.base.transform_primitive_base import TransformPrimitive
from featuretools.utils.gen_utils import Library


class AddNumeric(TransformPrimitive):
    """Performs element-wise addition of two lists.

    Description:
        Given a list of values X and a list of values
        Y, determine the sum of each value in X with its
        corresponding value in Y.

    Examples:
        >>> add_numeric = AddNumeric()
        >>> add_numeric([2, 1, 2], [1, 2, 2]).tolist()
        [3, 3, 4]
    """

    name = "add_numeric"
    input_types = [
        ColumnSchema(semantic_tags={"numeric"}),
        ColumnSchema(semantic_tags={"numeric"}),
    ]
    return_type = ColumnSchema(semantic_tags={"numeric"})
    commutative = True
    compatibility = [Library.PANDAS, Library.DASK, Library.SPARK]
    description_template = "the sum of {} and {}"

    def get_function(self):
        return np.add

    def generate_name(self, base_feature_names):
        return "%s + %s" % (base_feature_names[0], base_feature_names[1])



from datetime import date

import pandas as pd
from woodwork.column_schema import ColumnSchema
from woodwork.logical_types import Categorical, Datetime

from featuretools.primitives.base import TransformPrimitive


class Season(TransformPrimitive):
    """Determines the season of a given datetime.
        Returns winter, spring, summer, or fall.
        This only works for northern hemisphere.

    Description:
        Given a list of datetimes, return the season of each one
        (`winter`, `spring`, `summer`, or `fall`).

    Examples:
        >>> from datetime import datetime
        >>> times = [datetime(2019, 1, 1),
        ...          datetime(2019, 4, 15),
        ...          datetime(2019, 7, 20),
        ...          datetime(2019, 12, 30)]
        >>> season = Season()
        >>> season(times).tolist()
        ['winter', 'spring', 'summer', 'winter']
    """

    name = "season"
    input_types = [ColumnSchema(logical_type=Datetime)]
    return_type = ColumnSchema(logical_type=Categorical, semantic_tags={"category"})

    def get_function(self):
        def season(x):
            # https://stackoverflow.com/a/28688724/2512385
            Y = 2000  # dummy leap year to allow input X-02-29 (leap day)
            seasons = [
                ("winter", (date(Y, 1, 1), date(Y, 3, 20))),
                ("spring", (date(Y, 3, 21), date(Y, 6, 20))),
                ("summer", (date(Y, 6, 21), date(Y, 9, 22))),
                ("fall", (date(Y, 9, 23), date(Y, 12, 20))),
                ("winter", (date(Y, 12, 21), date(Y, 12, 31))),
            ]
            x = x.apply(lambda x: x.replace(year=2000))

            def get_season(dt):
                for season, (start, end) in seasons:
                    if not pd.isna(dt) and start <= dt.date() <= end:
                        return season
                return pd.NA

            new = x.apply(get_season).astype(dtype="string")
            return new

        return season



import numpy as np
from woodwork.column_schema import ColumnSchema
from woodwork.logical_types import Categorical, Datetime

from featuretools.primitives.base import TransformPrimitive


class DateToTimeZone(TransformPrimitive):
    """Determines the timezone of a datetime.

    Description:
        Given a list of datetimes, extract the timezone from each
        one. Looks for the `tzinfo` attribute on `datetime.datetime`
        objects. If the datetime has no timezone or the date is
        missing, return `NaN`.

    Examples:
        >>> from datetime import datetime
        >>> from pytz import timezone
        >>> date_to_time_zone = DateToTimeZone()
        >>> dates = [datetime(2010, 1, 1, tzinfo=timezone("America/Los_Angeles")),
        ...          datetime(2010, 1, 1, tzinfo=timezone("America/New_York")),
        ...          datetime(2010, 1, 1, tzinfo=timezone("America/Chicago")),
        ...          datetime(2010, 1, 1)]
        >>> date_to_time_zone(dates).tolist()
        ['America/Los_Angeles', 'America/New_York', 'America/Chicago', nan]
    """

    name = "date_to_time_zone"
    input_types = [ColumnSchema(logical_type=Datetime)]
    return_type = ColumnSchema(logical_type=Categorical, semantic_tags={"category"})

    def get_function(self):
        def date_to_time_zone(x):
            return x.apply(lambda x: x.tzinfo.zone if x.tzinfo else np.nan)

        return date_to_time_zone



from woodwork.column_schema import ColumnSchema

from featuretools.primitives.base import TransformPrimitive
from featuretools.utils.gen_utils import Library


class Negate(TransformPrimitive):
    """Negates a numeric value.

    Examples:
        >>> negate = Negate()
        >>> negate([1.0, 23.2, -7.0]).tolist()
        [-1.0, -23.2, 7.0]
    """

    name = "negate"
    input_types = [ColumnSchema(semantic_tags={"numeric"})]
    return_type = ColumnSchema(semantic_tags={"numeric"})
    compatibility = [Library.PANDAS, Library.DASK, Library.SPARK]
    description_template = "the negation of {}"

    def get_function(self):
        def negate(vals):
            return vals * -1

        return negate

    def generate_name(self, base_feature_names):
        return "-(%s)" % (base_feature_names[0])



from woodwork.column_schema import ColumnSchema

from featuretools.primitives.base import TransformPrimitive


class CumSum(TransformPrimitive):
    """Calculates the cumulative sum.

    Description:
        Given a list of values, return the cumulative sum
        (or running total). There is no set window, so the
        sum at each point is calculated over all prior values.
        `NaN` values will return `NaN`, but in the window of a
        cumulative caluclation, they're ignored.

    Examples:
        >>> cum_sum = CumSum()
        >>> cum_sum([1, 2, 3, 4, None, 5]).tolist()
        [1.0, 3.0, 6.0, 10.0, nan, 15.0]
    """

    name = "cum_sum"
    input_types = [ColumnSchema(semantic_tags={"numeric"})]
    return_type = ColumnSchema(semantic_tags={"numeric"})
    uses_full_dataframe = True
    description_template = "the cumulative sum of {}"

    def get_function(self):
        def cum_sum(values):
            return values.cumsum()

        return cum_sum



import numpy as np
import pandas as pd

from featuretools.primitives import (
    FullNameToFirstName,
    FullNameToLastName,
    FullNameToTitle,
)
from featuretools.tests.primitive_tests.utils import (
    PrimitiveTestBase,
    find_applicable_primitives,
    valid_dfs,
)


class TestFullNameToFirstName(PrimitiveTestBase):
    primitive = FullNameToFirstName

    def test_urls(self):
        # note this implementation incorrectly identifies the first
        # name for 'Oliva y Ocana, Dona. Fermina'
        primitive_func = self.primitive().get_function()
        names = pd.Series(
            [
                "Spector, Mr. Woolf",
                "Oliva y Ocana, Dona. Fermina",
                "Saether, Mr. Simon Sivertsen",
                "Ware, Mr. Frederick",
                "Peter, Master. Michael J",
            ],
        )
        answer = pd.Series(["Woolf", "Oliva", "Simon", "Frederick", "Michael"])
        pd.testing.assert_series_equal(primitive_func(names), answer, check_names=False)

    def test_no_title(self):
        primitive_func = self.primitive().get_function()
        names = pd.Series(
            [
                "Peter, Michael J",
                "James Masters",
                "Kate Elizabeth Brown-Jones",
            ],
        )
        answer = pd.Series(["Michael", "James", "Kate"], dtype=object)
        pd.testing.assert_series_equal(primitive_func(names), answer, check_names=False)

    def test_empty_string(self):
        primitive_func = self.primitive().get_function()
        names = pd.Series(
            [
                "Peter, Michael J",
                "",
                "Kate Elizabeth Brown-Jones",
            ],
        )
        answer = pd.Series(["Michael", np.nan, "Kate"], dtype=object)
        pd.testing.assert_series_equal(primitive_func(names), answer, check_names=False)

    def test_single_name(self):
        primitive_func = self.primitive().get_function()
        names = pd.Series(
            [
                "Peter, Michael J",
                "James",
                "Kate Elizabeth Brown-Jones",
            ],
        )
        answer = pd.Series(["Michael", "James", "Kate"], dtype=object)
        pd.testing.assert_series_equal(primitive_func(names), answer, check_names=False)

    def test_nan(self):
        primitive_func = self.primitive().get_function()
        names = pd.Series(["Mr. James Brown", np.nan, None])
        answer = pd.Series(["James", np.nan, np.nan])
        pd.testing.assert_series_equal(primitive_func(names), answer, check_names=False)

    def test_with_featuretools(self, pd_es):
        transform, aggregation = find_applicable_primitives(self.primitive)
        primitive_instance = self.primitive()
        transform.append(primitive_instance)
        valid_dfs(pd_es, aggregation, transform, self.primitive)


class TestFullNameToLastName(PrimitiveTestBase):
    primitive = FullNameToLastName

    def test_urls(self):
        primitive_func = self.primitive().get_function()
        names = pd.Series(
            [
                "Spector, Mr. Woolf",
                "Oliva y Ocana, Dona. Fermina",
                "Saether, Mr. Simon Sivertsen",
                "Ware, Mr. Frederick",
                "Peter, Master. Michael J",
            ],
        )
        answer = pd.Series(["Spector", "Oliva y Ocana", "Saether", "Ware", "Peter"])
        pd.testing.assert_series_equal(primitive_func(names), answer, check_names=False)

    def test_no_title(self):
        primitive_func = self.primitive().get_function()
        names = pd.Series(
            [
                "Peter, Michael J",
                "James Masters",
                "Kate Elizabeth Brown-Jones",
            ],
        )
        answer = pd.Series(["Peter", "Masters", "Brown-Jones"], dtype=object)
        pd.testing.assert_series_equal(primitive_func(names), answer, check_names=False)

    def test_empty_string(self):
        primitive_func = self.primitive().get_function()
        names = pd.Series(
            [
                "Peter, Michael J",
                "",
                "Kate Elizabeth Brown-Jones",
            ],
        )
        answer = pd.Series(["Peter", np.nan, "Brown-Jones"], dtype=object)
        pd.testing.assert_series_equal(primitive_func(names), answer, check_names=False)

    def test_single_name(self):
        primitive_func = self.primitive().get_function()
        names = pd.Series(
            [
                "Peter, Michael J",
                "James",
                "Kate Elizabeth Brown-Jones",
            ],
        )
        answer = pd.Series(["Peter", np.nan, "Brown-Jones"], dtype=object)
        pd.testing.assert_series_equal(primitive_func(names), answer, check_names=False)

    def test_nan(self):
        primitive_func = self.primitive().get_function()
        names = pd.Series(["Mr. James Brown", np.nan, None])
        answer = pd.Series(["Brown", np.nan, np.nan])
        pd.testing.assert_series_equal(primitive_func(names), answer, check_names=False)

    def test_with_featuretools(self, pd_es):
        transform, aggregation = find_applicable_primitives(self.primitive)
        primitive_instance = self.primitive()
        transform.append(primitive_instance)
        valid_dfs(pd_es, aggregation, transform, self.primitive)


class TestFullNameToTitle(PrimitiveTestBase):
    primitive = FullNameToTitle

    def test_urls(self):
        primitive_func = self.primitive().get_function()
        names = pd.Series(
            [
                "Spector, Mr. Woolf",
                "Oliva y Ocana, Dona. Fermina",
                "Saether, Mr. Simon Sivertsen",
                "Ware, Mr. Frederick",
                "Peter, Master. Michael J",
                "Mr. Brown",
            ],
        )
        answer = pd.Series(["Mr", "Dona", "Mr", "Mr", "Master", "Mr"])
        pd.testing.assert_series_equal(primitive_func(names), answer, check_names=False)

    def test_no_title(self):
        primitive_func = self.primitive().get_function()
        names = pd.Series(
            [
                "Peter, Michael J",
                "James Master.",
                "Mrs Brown",
                "",
            ],
        )
        answer = pd.Series([np.nan, np.nan, np.nan, np.nan], dtype=object)
        pd.testing.assert_series_equal(primitive_func(names), answer, check_names=False)

    def test_nan(self):
        primitive_func = self.primitive().get_function()
        names = pd.Series(["Mr. Brown", np.nan, None])
        answer = pd.Series(["Mr", np.nan, np.nan])
        pd.testing.assert_series_equal(primitive_func(names), answer, check_names=False)

    def test_with_featuretools(self, pd_es):
        transform, aggregation = find_applicable_primitives(self.primitive)
        primitive_instance = self.primitive()
        transform.append(primitive_instance)
        valid_dfs(pd_es, aggregation, transform, self.primitive)



import urllib.request

import pandas as pd
import pytest

from featuretools import EntitySet
from featuretools.demo import load_flight, load_mock_customer, load_retail, load_weather


@pytest.fixture(autouse=True)
def set_testing_headers():
    opener = urllib.request.build_opener()
    opener.addheaders = [("Testing", "True")]
    urllib.request.install_opener(opener)


def test_load_retail_diff():
    nrows = 10
    es_first = load_retail(nrows=nrows)
    assert isinstance(es_first, EntitySet)
    assert es_first["order_products"].shape[0] == nrows
    nrows_second = 11
    es_second = load_retail(nrows=nrows_second)
    assert es_second["order_products"].shape[0] == nrows_second


def test_mock_customer():
    n_customers = 4
    n_products = 3
    n_sessions = 30
    n_transactions = 400
    es = load_mock_customer(
        n_customers=n_customers,
        n_products=n_products,
        n_sessions=n_sessions,
        n_transactions=n_transactions,
        random_seed=0,
        return_entityset=True,
    )
    assert isinstance(es, EntitySet)
    df_names = [df.ww.name for df in es.dataframes]
    expected_names = ["transactions", "products", "sessions", "customers"]
    assert set(expected_names) == set(df_names)
    assert len(es["customers"]) == 4
    assert len(es["products"]) == 3
    assert len(es["sessions"]) == 30
    assert len(es["transactions"]) == 400


def test_load_flight():
    es = load_flight(
        month_filter=[1],
        categorical_filter={"origin_city": ["Charlotte, NC"]},
        return_single_table=False,
        nrows=1000,
    )
    assert isinstance(es, EntitySet)
    dataframe_names = ["airports", "flights", "trip_logs", "airlines"]
    realvals = [(11, 3), (13, 9), (103, 21), (1, 1)]
    for i, name in enumerate(dataframe_names):
        assert es[name].shape == realvals[i]


def test_weather():
    es = load_weather()
    assert isinstance(es, EntitySet)
    dataframe_names = ["temperatures"]
    realvals = [(3650, 3)]
    for i, name in enumerate(dataframe_names):
        assert es[name].shape == realvals[i]
    es = load_weather(return_single_table=True)
    assert isinstance(es, pd.DataFrame)



import numpy as np
import pandas as pd
import pytest

from featuretools.primitives import (
    RollingCount,
    RollingMax,
    RollingMean,
    RollingMin,
    RollingOutlierCount,
    RollingSTD,
    RollingTrend,
)
from featuretools.primitives.standard.transform.time_series.utils import (
    apply_rolling_agg_to_series,
)
from featuretools.tests.primitive_tests.utils import get_number_from_offset


@pytest.mark.parametrize(
    "window_length, gap",
    [
        (5, 2),
        (5, 0),
        ("5d", "7d"),
        ("5d", "0d"),
    ],
)
@pytest.mark.parametrize("min_periods", [1, 0, 2, 5])
def test_rolling_max(min_periods, window_length, gap, window_series_pd):
    gap_num = get_number_from_offset(gap)
    window_length_num = get_number_from_offset(window_length)
    # Since we're using a uniform series we can check correctness using numeric parameters
    expected_vals = apply_rolling_agg_to_series(
        window_series_pd,
        lambda x: x.max(),
        window_length_num,
        gap=gap_num,
        min_periods=min_periods,
    )

    primitive_instance = RollingMax(
        window_length=window_length,
        gap=gap,
        min_periods=min_periods,
    )
    primitive_func = primitive_instance.get_function()

    actual_vals = pd.Series(
        primitive_func(window_series_pd.index, pd.Series(window_series_pd.values)),
    )

    # Since min_periods of 0 is the same as min_periods of 1
    num_nans_from_min_periods = min_periods or 1

    assert actual_vals.isna().sum() == gap_num + num_nans_from_min_periods - 1
    pd.testing.assert_series_equal(pd.Series(expected_vals), actual_vals)


@pytest.mark.parametrize(
    "window_length, gap",
    [
        (5, 2),
        (5, 0),
        ("5d", "7d"),
        ("5d", "0d"),
    ],
)
@pytest.mark.parametrize("min_periods", [1, 0, 2, 5])
def test_rolling_min(min_periods, window_length, gap, window_series_pd):
    gap_num = get_number_from_offset(gap)
    window_length_num = get_number_from_offset(window_length)

    # Since we're using a uniform series we can check correctness using numeric parameters
    expected_vals = apply_rolling_agg_to_series(
        window_series_pd,
        lambda x: x.min(),
        window_length_num,
        gap=gap_num,
        min_periods=min_periods,
    )

    primitive_instance = RollingMin(
        window_length=window_length,
        gap=gap,
        min_periods=min_periods,
    )
    primitive_func = primitive_instance.get_function()

    actual_vals = pd.Series(
        primitive_func(window_series_pd.index, pd.Series(window_series_pd.values)),
    )

    # Since min_periods of 0 is the same as min_periods of 1
    num_nans_from_min_periods = min_periods or 1

    assert actual_vals.isna().sum() == gap_num + num_nans_from_min_periods - 1
    pd.testing.assert_series_equal(pd.Series(expected_vals), actual_vals)


@pytest.mark.parametrize(
    "window_length, gap",
    [
        (5, 2),
        (5, 0),
        ("5d", "7d"),
        ("5d", "0d"),
    ],
)
@pytest.mark.parametrize("min_periods", [1, 0, 2, 5])
def test_rolling_mean(min_periods, window_length, gap, window_series_pd):
    gap_num = get_number_from_offset(gap)
    window_length_num = get_number_from_offset(window_length)

    # Since we're using a uniform series we can check correctness using numeric parameters
    expected_vals = apply_rolling_agg_to_series(
        window_series_pd,
        np.mean,
        window_length_num,
        gap=gap_num,
        min_periods=min_periods,
    )

    primitive_instance = RollingMean(
        window_length=window_length,
        gap=gap,
        min_periods=min_periods,
    )
    primitive_func = primitive_instance.get_function()

    actual_vals = pd.Series(
        primitive_func(window_series_pd.index, pd.Series(window_series_pd.values)),
    )

    # Since min_periods of 0 is the same as min_periods of 1
    num_nans_from_min_periods = min_periods or 1

    assert actual_vals.isna().sum() == gap_num + num_nans_from_min_periods - 1
    pd.testing.assert_series_equal(pd.Series(expected_vals), actual_vals)


@pytest.mark.parametrize(
    "window_length, gap",
    [
        (5, 2),
        (5, 0),
        ("5d", "7d"),
        ("5d", "0d"),
    ],
)
@pytest.mark.parametrize("min_periods", [1, 0, 2, 5])
def test_rolling_std(min_periods, window_length, gap, window_series_pd):
    gap_num = get_number_from_offset(gap)
    window_length_num = get_number_from_offset(window_length)

    # Since we're using a uniform series we can check correctness using numeric parameters
    expected_vals = apply_rolling_agg_to_series(
        window_series_pd,
        lambda x: x.std(),
        window_length_num,
        gap=gap_num,
        min_periods=min_periods,
    )

    primitive_instance = RollingSTD(
        window_length=window_length,
        gap=gap,
        min_periods=min_periods,
    )
    primitive_func = primitive_instance.get_function()

    actual_vals = pd.Series(
        primitive_func(window_series_pd.index, pd.Series(window_series_pd.values)),
    )

    # Since min_periods of 0 is the same as min_periods of 1
    num_nans_from_min_periods = min_periods or 2

    if min_periods in [0, 1]:
        # the additional nan is because std pandas function returns NaN if there's only one value
        num_nans = gap_num + 1
    else:
        num_nans = gap_num + num_nans_from_min_periods - 1

    # The extra 1 at the beginning is because the std pandas function returns NaN if there's only one value
    assert actual_vals.isna().sum() == num_nans
    pd.testing.assert_series_equal(pd.Series(expected_vals), actual_vals)


@pytest.mark.parametrize(
    "window_length, gap",
    [
        (5, 2),
        ("6d", "7d"),
    ],
)
def test_rolling_count(window_length, gap, window_series_pd):
    gap_num = get_number_from_offset(gap)
    window_length_num = get_number_from_offset(window_length)

    expected_vals = apply_rolling_agg_to_series(
        window_series_pd,
        lambda x: x.count(),
        window_length_num,
        gap=gap_num,
    )

    primitive_instance = RollingCount(
        window_length=window_length,
        gap=gap,
        min_periods=window_length_num,
    )
    primitive_func = primitive_instance.get_function()

    actual_vals = pd.Series(primitive_func(window_series_pd.index))

    num_nans = gap_num + window_length_num - 1
    assert actual_vals.isna().sum() == num_nans
    # RollingCount will not match the exact roll_series_with_gap call,
    # because it handles the min_periods difference within the primitive
    pd.testing.assert_series_equal(
        pd.Series(expected_vals).iloc[num_nans:],
        actual_vals.iloc[num_nans:],
    )


@pytest.mark.parametrize(
    "min_periods, expected_num_nams",
    [(0, 2), (1, 2), (3, 4), (5, 6)],  # 0 and 1 get treated the same
)
@pytest.mark.parametrize("window_length, gap", [("5d", "2d"), (5, 2)])
def test_rolling_count_primitive_min_periods_nans(
    window_length,
    gap,
    min_periods,
    expected_num_nams,
    window_series_pd,
):
    primitive_instance = RollingCount(
        window_length=window_length,
        gap=gap,
        min_periods=min_periods,
    )
    primitive_func = primitive_instance.get_function()
    vals = pd.Series(primitive_func(window_series_pd.index))

    assert vals.isna().sum() == expected_num_nams


@pytest.mark.parametrize(
    "min_periods, expected_num_nams",
    [(0, 0), (1, 0), (3, 2), (5, 4)],  # 0 and 1 get treated the same
)
@pytest.mark.parametrize("window_length, gap", [("5d", "0d"), (5, 0)])
def test_rolling_count_with_no_gap(
    window_length,
    gap,
    min_periods,
    expected_num_nams,
    window_series_pd,
):
    primitive_instance = RollingCount(
        window_length=window_length,
        gap=gap,
        min_periods=min_periods,
    )
    primitive_func = primitive_instance.get_function()
    vals = pd.Series(primitive_func(window_series_pd.index))

    assert vals.isna().sum() == expected_num_nams


@pytest.mark.parametrize(
    "window_length, gap, expected_vals",
    [
        (3, 0, [np.nan, np.nan, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1]),
        (
            4,
            1,
            [np.nan, np.nan, np.nan, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1],
        ),
        (
            "5d",
            "7d",
            [
                np.nan,
                np.nan,
                np.nan,
                np.nan,
                np.nan,
                np.nan,
                np.nan,
                np.nan,
                np.nan,
                1,
                1,
                1,
                1,
                1,
                1,
                1,
                1,
                1,
                1,
                1,
            ],
        ),
        (
            "5d",
            "0d",
            [np.nan, np.nan, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1],
        ),
    ],
)
def test_rolling_trend(window_length, gap, expected_vals, window_series_pd):
    primitive_instance = RollingTrend(window_length=window_length, gap=gap)

    actual_vals = primitive_instance(window_series_pd.index, window_series_pd.values)

    pd.testing.assert_series_equal(pd.Series(expected_vals), pd.Series(actual_vals))


def test_rolling_trend_window_length_less_than_three(window_series_pd):
    primitive_instance = RollingTrend(window_length=2)

    vals = primitive_instance(window_series_pd.index, window_series_pd.values)

    for v in vals:
        assert np.isnan(v)


@pytest.mark.parametrize(
    "primitive",
    [
        RollingCount,
        RollingMax,
        RollingMin,
        RollingMean,
        RollingOutlierCount,
    ],
)
def test_rolling_primitives_non_uniform(primitive):
    # When the data isn't uniform, this impacts the number of values in each rolling window
    datetimes = (
        list(pd.date_range(start="2017-01-01", freq="1d", periods=3))
        + list(pd.date_range(start="2017-01-10", freq="2d", periods=4))
        + list(pd.date_range(start="2017-01-22", freq="1d", periods=7))
    )
    no_freq_series = pd.Series(range(len(datetimes)), index=datetimes)

    # Should match RollingCount exactly and have same nan values as other primitives
    expected_series = pd.Series(
        [None, 1, 2] + [None, 1, 1, 1] + [None, 1, 2, 3, 3, 3, 3],
    )

    primitive_instance = primitive(window_length="3d", gap="1d")
    if isinstance(primitive_instance, RollingCount):
        rolled_series = pd.Series(primitive_instance(no_freq_series.index))
        pd.testing.assert_series_equal(rolled_series, expected_series)
    else:
        rolled_series = pd.Series(
            primitive_instance(no_freq_series.index, pd.Series(no_freq_series.values)),
        )
        pd.testing.assert_series_equal(expected_series.isna(), rolled_series.isna())


def test_rolling_std_non_uniform():
    # When the data isn't uniform, this impacts the number of values in each rolling window
    datetimes = (
        list(pd.date_range(start="2017-01-01", freq="1d", periods=3))
        + list(pd.date_range(start="2017-01-10", freq="2d", periods=4))
        + list(pd.date_range(start="2017-01-22", freq="1d", periods=7))
    )
    no_freq_series = pd.Series(range(len(datetimes)), index=datetimes)

    # There will be at least two null values at the beginning of each range's rows, the first for the
    # row skipped by the gap, and the second because pandas' std returns NaN if there's only one row
    expected_series = pd.Series(
        [None, None, 0.707107]
        + [None, None, None, None]
        + [  # Because the freq was 2 days, there will never be more than 1 observation
            None,
            None,
            0.707107,
            1.0,
            1.0,
            1.0,
            1.0,
        ],
    )

    primitive_instance = RollingSTD(window_length="3d", gap="1d")
    rolled_series = pd.Series(
        primitive_instance(no_freq_series.index, pd.Series(no_freq_series.values)),
    )

    pd.testing.assert_series_equal(rolled_series, expected_series)


def test_rolling_trend_non_uniform():
    datetimes = (
        list(pd.date_range(start="2017-01-01", freq="1d", periods=3))
        + list(pd.date_range(start="2017-01-10", freq="2d", periods=4))
        + list(pd.date_range(start="2017-01-22", freq="1d", periods=7))
    )
    no_freq_series = pd.Series(range(len(datetimes)), index=datetimes)
    expected_series = pd.Series(
        [None, None, None]
        + [None, None, None, None]
        + [
            None,
            None,
            None,
            1.0,
            1.0,
            1.0,
            1.0,
        ],
    )
    primitive_instance = RollingTrend(window_length="3d", gap="1d")
    rolled_series = pd.Series(
        primitive_instance(no_freq_series.index, pd.Series(no_freq_series.values)),
    )
    pd.testing.assert_series_equal(rolled_series, expected_series)


@pytest.mark.parametrize(
    "window_length, gap",
    [
        (5, 2),
        (5, 0),
        ("5d", "7d"),
        ("5d", "0d"),
    ],
)
@pytest.mark.parametrize(
    "min_periods",
    [1, 0, 2, 5],
)
def test_rolling_outlier_count(
    min_periods,
    window_length,
    gap,
    rolling_outlier_series_pd,
):
    primitive_instance = RollingOutlierCount(
        window_length=window_length,
        gap=gap,
        min_periods=min_periods,
    )

    primitive_func = primitive_instance.get_function()

    actual_vals = pd.Series(
        primitive_func(
            rolling_outlier_series_pd.index,
            pd.Series(rolling_outlier_series_pd.values),
        ),
    )

    expected_vals = apply_rolling_agg_to_series(
        series=rolling_outlier_series_pd,
        agg_func=primitive_instance.get_outliers_count,
        window_length=window_length,
        gap=gap,
        min_periods=min_periods,
    )

    # Since min_periods of 0 is the same as min_periods of 1
    num_nans_from_min_periods = min_periods or 1
    assert (
        actual_vals.isna().sum()
        == get_number_from_offset(gap) + num_nans_from_min_periods - 1
    )
    pd.testing.assert_series_equal(actual_vals, pd.Series(data=expected_vals))



import pandas as pd
import pytest

from featuretools import calculate_feature_matrix, dfs, list_primitives
from featuretools.feature_base.cache import feature_cache
from featuretools.primitives import get_aggregation_primitives, get_transform_primitives
from featuretools.tests.testing_utils import to_pandas
from featuretools.utils.gen_utils import Library

UNSUPPORTED = [
    p.name
    for p in get_transform_primitives().values()
    if Library.DASK not in p.compatibility
]
UNSUPPORTED += [
    p.name
    for p in get_aggregation_primitives().values()
    if Library.DASK not in p.compatibility
]


@pytest.fixture(autouse=True)
def reset_dfs_cache():
    feature_cache.enabled = False
    feature_cache.clear_all()


def test_transform(pd_es, dask_es):
    pytest.skip(
        "TODO: Dask issue with `series.eq`. Fix once Dask Issue #7957 is closed.",
    )
    primitives = list_primitives()
    trans_list = primitives[primitives["type"] == "transform"]["name"].tolist()
    trans_primitives = [prim for prim in trans_list if prim not in UNSUPPORTED]
    agg_primitives = []
    cutoff_time = pd.Timestamp("2019-01-05 04:00")

    assert pd_es == dask_es

    # Run DFS using each dataframe as a target and confirm results match
    for df in pd_es.dataframes:
        features = dfs(
            entityset=pd_es,
            target_dataframe_name=df.ww.name,
            trans_primitives=trans_primitives,
            agg_primitives=agg_primitives,
            max_depth=2,
            features_only=True,
        )

        dask_features = dfs(
            entityset=dask_es,
            target_dataframe_name=df.ww.name,
            trans_primitives=trans_primitives,
            agg_primitives=agg_primitives,
            max_depth=2,
            features_only=True,
        )
        assert features == dask_features

        # Calculate feature matrix values to confirm output is the same between dask and pandas.
        # Not testing on all returned features due to long run times.
        fm = calculate_feature_matrix(
            features=features[:100],
            entityset=pd_es,
            cutoff_time=cutoff_time,
        )
        dask_fm = calculate_feature_matrix(
            features=dask_features[:100],
            entityset=dask_es,
            cutoff_time=cutoff_time,
        )

        # Categorical categories can be ordered differently, this makes sure they are the same
        dask_fm = dask_fm.astype(fm.dtypes)

        # Use the same columns and make sure both indexes are sorted the same
        dask_computed_fm = (
            dask_fm.compute().set_index(df.ww.index).loc[fm.index][fm.columns]
        )
        pd.testing.assert_frame_equal(fm, dask_computed_fm)


def test_aggregation(pd_es, dask_es):
    primitives = list_primitives()
    trans_primitives = []
    agg_list = primitives[primitives["type"] == "aggregation"]["name"].tolist()
    agg_primitives = [prim for prim in agg_list if prim not in UNSUPPORTED]

    assert pd_es == dask_es

    # Run DFS using each dataframe as a target and confirm results match
    for df in pd_es.dataframes:
        fm, _ = dfs(
            entityset=pd_es,
            target_dataframe_name=df.ww.name,
            trans_primitives=trans_primitives,
            agg_primitives=agg_primitives,
            cutoff_time=pd.Timestamp("2019-01-05 04:00"),
            max_depth=2,
        )

        dask_fm, _ = dfs(
            entityset=dask_es,
            target_dataframe_name=df.ww.name,
            trans_primitives=trans_primitives,
            agg_primitives=agg_primitives,
            cutoff_time=pd.Timestamp("2019-01-05 04:00"),
            max_depth=2,
        )

        # Categorical categories can be ordered differently, this makes sure they
        # are the same, including the index column
        index_col = df.ww.index
        fm = fm.reset_index()
        dask_fm = dask_fm.astype(fm.dtypes)
        fm = fm.set_index(index_col)

        pd.testing.assert_frame_equal(
            fm.sort_index(),
            to_pandas(dask_fm, index=index_col, sort_index=True),
        )



import numpy as np
import pandas as pd
from woodwork.column_schema import ColumnSchema
from woodwork.logical_types import Datetime, Double

from featuretools.primitives.base import TransformPrimitive


class NthWeekOfMonth(TransformPrimitive):
    """Determines the nth week of the month from a given date.

    Description:
        Converts a datetime to an float representing the week
        of the month in which the date falls. The first day of
        the month starts week 1, and the week number is incremented
        each Sunday.

    Examples:
        >>> from datetime import datetime
        >>> nth_week_of_month = NthWeekOfMonth()
        >>> times = [datetime(2019, 3, 1),
        ...          datetime(2019, 3, 3),
        ...          datetime(2019, 3, 31),
        ...          datetime(2019, 3, 30)]
        >>> nth_week_of_month(times).tolist()
        [1.0, 2.0, 6.0, 5.0]
    """

    name = "nth_week_of_month"
    input_types = [ColumnSchema(logical_type=Datetime)]
    return_type = ColumnSchema(logical_type=Double, semantic_tags={"numeric"})

    def get_function(self):
        def nth_week_of_month(x):
            df = pd.DataFrame({"date": x})
            df["first_day"] = df.date - pd.to_timedelta(df["date"].dt.day - 1, unit="d")
            df["dom"] = df.date.dt.day
            df["first_day_weekday"] = df.first_day.dt.weekday
            df["adjusted_dom"] = df.dom + df.first_day_weekday + 1
            df.loc[df["first_day_weekday"].astype(float) == 6.0, "adjusted_dom"] = df[
                "dom"
            ]
            df["week_of_month"] = np.ceil(df.adjusted_dom / 7.0)
            return df.week_of_month.values

        return nth_week_of_month



import numpy as np
import pandas as pd

from featuretools.primitives import NumberOfMentions
from featuretools.tests.primitive_tests.utils import (
    PrimitiveTestBase,
    find_applicable_primitives,
    valid_dfs,
)


class TestNumberOfMentions(PrimitiveTestBase):
    primitive = NumberOfMentions

    def test_regular_input(self):
        x = pd.Series(
            [
                "@hello @hi @hello",
                "@and@",
                "andorandorand",
            ],
        )
        expected = [3.0, 0.0, 0.0]
        actual = self.primitive().get_function()(x)
        np.testing.assert_array_equal(actual, expected)

    def test_unicode_input(self):
        x = pd.Series(
            [
                "@Ángel @Æ @ĘÁÊÚ",
                "@@@@Āndandandandand@",
                "andorandorand @32309",
                "example@gmail.com",
                "@example-20329",
            ],
        )
        expected = [3.0, 0.0, 1.0, 0.0, 1.0]
        actual = self.primitive().get_function()(x)
        np.testing.assert_array_equal(actual, expected)

    def test_multiline(self):
        x = pd.Series(
            [
                "@\n\t\n",
                "@mention\n @mention2\n@\n\n",
            ],
        )

        expected = [0.0, 2.0]
        actual = self.primitive().get_function()(x)
        np.testing.assert_array_equal(actual, expected)

    def test_null(self):
        x = pd.Series([np.nan, pd.NA, None, "@test"])

        actual = self.primitive().get_function()(x)
        expected = [np.nan, np.nan, np.nan, 1.0]
        np.testing.assert_array_equal(actual, expected)

    def test_alphanumeric_and_special(self):
        x = pd.Series(["@1or0", "@12", "#??!>@?@#>"])

        actual = self.primitive().get_function()(x)
        expected = [1.0, 1.0, 0.0]
        np.testing.assert_array_equal(actual, expected)

    def test_underscore(self):
        x = pd.Series(["@user1", "@__yes", "#??!>@?@#>"])

        actual = self.primitive().get_function()(x)
        expected = [1.0, 1.0, 0.0]
        np.testing.assert_array_equal(actual, expected)

    def test_with_featuretools(self, es):
        transform, aggregation = find_applicable_primitives(self.primitive)
        primitive_instance = self.primitive()
        transform.append(primitive_instance)
        valid_dfs(es, aggregation, transform, self.primitive)



from __future__ import annotations

import hashlib
from itertools import combinations
from typing import Any, Dict, List, Optional, Set, Type, Union, cast

from woodwork.logical_types import LogicalType

from featuretools.feature_discovery.LiteFeature import LiteFeature
from featuretools.feature_discovery.type_defs import ANY
from featuretools.feature_discovery.utils import hash_primitive, logical_types_map
from featuretools.primitives.base.primitive_base import PrimitiveBase
from featuretools.primitives.utils import (
    PrimitivesDeserializer,
)


class FeatureCollection:
    def __init__(self, features: List[LiteFeature]):
        self._all_features: List[LiteFeature] = features
        self.indexed = False
        self.sorted = False
        self._hash_key: Optional[str] = None

    def sort_features(self):
        if not self.sorted:
            self._all_features = sorted(self._all_features)
            self.sorted = True

    def __repr__(self):
        return f"<FeatureCollection ({self.hash_key[:5]}) n_features={len(self._all_features)} indexed={self.indexed}>"

    @property
    def all_features(self):
        return self._all_features.copy()

    @property
    def hash_key(self) -> str:
        if self._hash_key is None:
            if not self.sorted:
                self.sort_features()
            self._set_hash()
        assert self._hash_key is not None
        return self._hash_key

    def _set_hash(self):
        hash_msg = hashlib.sha256()

        for feature in self._all_features:
            hash_msg.update(feature.id.encode("utf-8"))

        self._hash_key = hash_msg.hexdigest()
        return self

    def __hash__(self):
        return hash(self.hash_key)

    def __eq__(self, other: FeatureCollection) -> bool:
        return self.hash_key == other.hash_key

    def reindex(self) -> FeatureCollection:
        self.by_logical_type: Dict[
            Union[Type[LogicalType], None],
            Set[LiteFeature],
        ] = {}
        self.by_tag: Dict[str, Set[LiteFeature]] = {}
        self.by_origin_feature: Dict[LiteFeature, Set[LiteFeature]] = {}
        self.by_depth: Dict[int, Set[LiteFeature]] = {}
        self.by_name: Dict[str, LiteFeature] = {}
        self.by_key: Dict[str, List[LiteFeature]] = {}

        for feature in self._all_features:
            for key in self.feature_to_keys(feature):
                self.by_key.setdefault(key, []).append(feature)

            logical_type = feature.logical_type
            self.by_logical_type.setdefault(logical_type, set()).add(feature)

            tags = feature.tags
            for tag in tags:
                self.by_tag.setdefault(tag, set()).add(feature)

            origin_features = feature.get_origin_features()
            for origin_feature in origin_features:
                self.by_origin_feature.setdefault(origin_feature, set()).add(feature)

            if feature.depth == 0:
                self.by_origin_feature.setdefault(feature, set()).add(feature)

            feature_name = feature.name
            assert feature_name is not None
            assert feature_name not in self.by_name

            self.by_name[feature_name] = feature

        self.indexed = True

        return self

    def get_by_logical_type(self, logical_type: Type[LogicalType]) -> Set[LiteFeature]:
        return self.by_logical_type.get(logical_type, set())

    def get_by_tag(self, tag: str) -> Set[LiteFeature]:
        return self.by_tag.get(tag, set())

    def get_by_origin_feature(self, origin_feature: LiteFeature) -> Set[LiteFeature]:
        return self.by_origin_feature.get(origin_feature, set())

    def get_by_origin_feature_name(self, name: str) -> Union[LiteFeature, None]:
        feature = self.by_name.get(name)
        return feature

    def get_dependencies_by_origin_name(self, name) -> Set[LiteFeature]:
        origin_feature = self.by_name.get(name)
        if origin_feature:
            return self.by_origin_feature[origin_feature]
        return set()

    def get_by_key(self, key: str) -> List[LiteFeature]:
        return self.by_key.get(key, [])

    def flatten_features(self) -> Dict[str, LiteFeature]:
        all_features_dict: Dict[str, LiteFeature] = {}

        def rfunc(feature_list: List[LiteFeature]):
            for feature in feature_list:
                all_features_dict.setdefault(feature.id, feature)
                rfunc(feature.base_features)

        rfunc(self._all_features)
        return all_features_dict

    def flatten_primitives(self) -> Dict[str, Dict[str, Any]]:
        all_primitives_dict: Dict[str, Dict[str, Any]] = {}

        def rfunc(feature_list: List[LiteFeature]):
            for feature in feature_list:
                if feature.primitive:
                    key, prim_dict = hash_primitive(feature.primitive)
                    all_primitives_dict.setdefault(key, prim_dict)
                rfunc(feature.base_features)

        rfunc(self._all_features)
        return all_primitives_dict

    def to_dict(self):
        all_primitives_dict = self.flatten_primitives()
        all_features_dict = self.flatten_features()

        return {
            "primitives": all_primitives_dict,
            "feature_ids": [f.id for f in self._all_features],
            "all_features": {k: f.to_dict() for k, f in all_features_dict.items()},
        }

    @staticmethod
    def feature_to_keys(feature: LiteFeature) -> List[str]:
        """
        Generate hashing keys from LiteFeature. For example:
        - LiteFeature("f1", Double, {"numeric"}) -> ['Double', 'numeric', 'Double,numeric', 'ANY']
        - LiteFeature("f1", Datetime, {"time_index"}) -> ['Datetime', 'time_index', 'Datetime,time_index', 'ANY']
        - LiteFeature("f1", Double, {"index", "other"}) -> ['Double', 'index', 'other', 'Double,index', 'Double,other', 'ANY']

                Args:
            feature (LiteFeature):

        Returns:
            List[str]
                List of hashing keys
        """
        keys: List[str] = []
        logical_type = feature.logical_type
        logical_type_name = None
        if logical_type is not None:
            logical_type_name = logical_type.__name__
            keys.append(logical_type_name)

        all_tags = sorted(feature.tags)

        tag_combinations = []

        # generate combinations of all lengths from 1 to the length of the input list
        for i in range(1, len(all_tags) + 1):
            # generate combinations of length i and append to the combinations_list
            for comb in combinations(all_tags, i):
                tag_combinations.append(list(comb))

        for tag_combination in tag_combinations:
            tags_key = ",".join(tag_combination)
            keys.append(tags_key)
            if logical_type_name:
                keys.append(f"{logical_type_name},{tags_key}")

        keys.append(ANY)
        return keys

    @staticmethod
    def from_dict(input_dict):
        primitive_deserializer = PrimitivesDeserializer()

        primitives = {}
        for prim_key, prim_dict in input_dict["primitives"].items():
            primitive = primitive_deserializer.deserialize_primitive(
                prim_dict,
            )
            assert isinstance(primitive, PrimitiveBase)
            primitives[prim_key] = primitive

        hydrated_features: Dict[str, LiteFeature] = {}

        feature_ids: List[str] = cast(List[str], input_dict["feature_ids"])
        all_features: Dict[str, Any] = cast(Dict[str, Any], input_dict["all_features"])

        def hydrate_feature(feature_id: str) -> LiteFeature:
            if feature_id in hydrated_features:
                return hydrated_features[feature_id]

            feature_dict = all_features[feature_id]
            base_features = [hydrate_feature(x) for x in feature_dict["base_features"]]

            logical_type = (
                logical_types_map[feature_dict["logical_type"]]
                if feature_dict["logical_type"]
                else None
            )

            hydrated_feature = LiteFeature(
                name=feature_dict["name"],
                logical_type=logical_type,
                tags=set(feature_dict["tags"]),
                primitive=primitives[feature_dict["primitive"]]
                if feature_dict["primitive"]
                else None,
                base_features=base_features,
                df_id=feature_dict["df_id"],
                related_features=set(),
                idx=feature_dict["idx"],
            )

            assert hydrated_feature.id == feature_dict["id"] == feature_id
            hydrated_features[feature_id] = hydrated_feature

            # need to link after features are stored on cache
            related_features = [
                hydrate_feature(x) for x in feature_dict["related_features"]
            ]
            hydrated_feature.related_features = set(related_features)

            return hydrated_feature

        return FeatureCollection([hydrate_feature(x) for x in feature_ids])



import logging

import pandas as pd

from featuretools.computational_backends.utils import get_ww_types_from_features
from featuretools.utils.gen_utils import make_tqdm_iterator

logger = logging.getLogger("featuretools")

DEFAULT_TOP_N = 10


def encode_features(
    feature_matrix,
    features,
    top_n=DEFAULT_TOP_N,
    include_unknown=True,
    to_encode=None,
    inplace=False,
    drop_first=False,
    verbose=False,
):
    """Encode categorical features

    Args:
        feature_matrix (pd.DataFrame): Dataframe of features.
        features (list[PrimitiveBase]): Feature definitions in feature_matrix.
        top_n (int or dict[string -> int]): Number of top values to include.
            If dict[string -> int] is used, key is feature name and value is
            the number of top values to include for that feature.
            If a feature's name is not in dictionary, a default value of 10 is used.
        include_unknown (pd.DataFrame): Add feature encoding an unknown class.
            defaults to True
        to_encode (list[str]): List of feature names to encode.
            features not in this list are unencoded in the output matrix
            defaults to encode all necessary features.
        inplace (bool): Encode feature_matrix in place. Defaults to False.
        drop_first (bool): Whether to get k-1 dummies out of k categorical
                levels by removing the first level.
                defaults to False
        verbose (str): Print progress info.

    Returns:
        (pd.Dataframe, list) : encoded feature_matrix, encoded features

    Example:
        .. ipython:: python
            :suppress:

            from featuretools.tests.testing_utils import make_ecommerce_entityset
            import featuretools as ft
            es = make_ecommerce_entityset()

        .. ipython:: python

            f1 = ft.Feature(es["log"].ww["product_id"])
            f2 = ft.Feature(es["log"].ww["purchased"])
            f3 = ft.Feature(es["log"].ww["value"])

            features = [f1, f2, f3]
            ids = [0, 1, 2, 3, 4, 5]
            feature_matrix = ft.calculate_feature_matrix(features, es,
                                                         instance_ids=ids)

            fm_encoded, f_encoded = ft.encode_features(feature_matrix,
                                                       features)
            f_encoded

            fm_encoded, f_encoded = ft.encode_features(feature_matrix,
                                                       features, top_n=2)
            f_encoded

            fm_encoded, f_encoded = ft.encode_features(feature_matrix, features,
                                                       include_unknown=False)
            f_encoded

            fm_encoded, f_encoded = ft.encode_features(feature_matrix, features,
                                                       to_encode=['purchased'])
            f_encoded

            fm_encoded, f_encoded = ft.encode_features(feature_matrix, features,
                                                       drop_first=True)
            f_encoded
    """
    if not isinstance(feature_matrix, pd.DataFrame):
        msg = "feature_matrix must be a Pandas DataFrame"
        raise TypeError(msg)

    if inplace:
        X = feature_matrix
    else:
        X = feature_matrix.copy()

    old_feature_names = set()
    for feature in features:
        for fname in feature.get_feature_names():
            assert fname in X.columns, "Feature %s not found in feature matrix" % (
                fname
            )
            old_feature_names.add(fname)

    pass_through = [col for col in X.columns if col not in old_feature_names]

    if verbose:
        iterator = make_tqdm_iterator(
            iterable=features,
            total=len(features),
            desc="Encoding pass 1",
            unit="feature",
        )
    else:
        iterator = features

    new_feature_list = []
    kept_columns = []
    encoded_columns = []
    columns_info = feature_matrix.ww.columns

    for f in iterator:
        # TODO: features with multiple columns are not encoded by this method,
        # which can cause an "encoded" matrix with non-numeric values
        is_discrete = {"category", "foreign_key"}.intersection(
            f.column_schema.semantic_tags,
        )
        if f.number_output_features > 1 or not is_discrete:
            if f.number_output_features > 1:
                logger.warning(
                    "Feature %s has multiple columns and will not "
                    "be encoded.  This may result in a matrix with"
                    " non-numeric values." % (f),
                )
            new_feature_list.append(f)
            kept_columns.extend(f.get_feature_names())
            continue

        if to_encode is not None and f.get_name() not in to_encode:
            new_feature_list.append(f)
            kept_columns.extend(f.get_feature_names())
            continue

        val_counts = X[f.get_name()].value_counts()
        # Remove 0 count category values
        val_counts = val_counts[val_counts > 0].to_frame()
        index_name = val_counts.index.name
        val_counts = val_counts.rename(columns={val_counts.columns[0]: "count"})
        if index_name is None:
            if "index" in val_counts.columns:
                index_name = "level_0"
            else:
                index_name = "index"
        val_counts.reset_index(inplace=True)
        val_counts = val_counts.sort_values(["count", index_name], ascending=False)
        val_counts.set_index(index_name, inplace=True)
        select_n = top_n
        if isinstance(top_n, dict):
            select_n = top_n.get(f.get_name(), DEFAULT_TOP_N)
        if drop_first:
            select_n = min(len(val_counts), top_n)
            select_n = max(select_n - 1, 1)
        unique = val_counts.head(select_n).index.tolist()
        for label in unique:
            add = f == label
            add_name = add.get_name()
            new_feature_list.append(add)
            new_col = X[f.get_name()] == label
            new_col.rename(add_name, inplace=True)
            encoded_columns.append(new_col)

        if include_unknown:
            unknown = f.isin(unique).NOT().rename(f.get_name() + " is unknown")
            unknown_name = unknown.get_name()
            new_feature_list.append(unknown)
            new_col = ~X[f.get_name()].isin(unique)
            new_col.rename(unknown_name, inplace=True)
            encoded_columns.append(new_col)

        if inplace:
            X.drop(f.get_name(), axis=1, inplace=True)

    kept_columns.extend(pass_through)

    if inplace:
        for encoded_column in encoded_columns:
            X[encoded_column.name] = encoded_column
    else:
        X = pd.concat([X[kept_columns]] + encoded_columns, axis=1)

    entityset = new_feature_list[0].entityset
    ww_init_kwargs = get_ww_types_from_features(new_feature_list, entityset)

    # Grab ww metadata from feature matrix since it may be more exact
    for column in kept_columns:
        ww_init_kwargs["logical_types"][column] = columns_info[column].logical_type
        ww_init_kwargs["semantic_tags"][column] = columns_info[column].semantic_tags
        ww_init_kwargs["column_origins"][column] = columns_info[column].origin

    X.ww.init(**ww_init_kwargs)
    return X, new_feature_list



import numpy as np
import pandas as pd

from featuretools.primitives import NumberOfUniqueWords
from featuretools.tests.primitive_tests.utils import (
    PrimitiveTestBase,
    find_applicable_primitives,
    valid_dfs,
)


class TestNumberOfUniqueWords(PrimitiveTestBase):
    primitive = NumberOfUniqueWords

    def test_general(self):
        x = pd.Series(
            [
                "test test test test",
                "test TEST test TEST",
                "and subsequent lines...",
            ],
        )

        expected = pd.Series([1, 2, 3])
        actual = self.primitive().get_function()(x)
        pd.testing.assert_series_equal(actual, expected, check_names=False)

    def test_special_characters_and_whitespace(self):
        x = pd.Series(["50% 50 50% \t\t\t\n\n", "a test* test"])

        expected = pd.Series([1, 2])
        actual = self.primitive().get_function()(x)
        pd.testing.assert_series_equal(actual, expected, check_names=False)

    def test_unicode_input(self):
        x = pd.Series(
            [
                "Ángel Angel Ángel ángel",
            ],
        )

        expected = pd.Series([3])
        actual = self.primitive().get_function()(x)
        pd.testing.assert_series_equal(actual, expected, check_names=False)

    def test_contractions(self):
        x = pd.Series(
            [
                "can't won't don't can't aren't won't don't they'd there's",
            ],
        )

        expected = pd.Series([6])
        actual = self.primitive().get_function()(x)
        pd.testing.assert_series_equal(actual, expected, check_names=False)

    def test_multiline(self):
        x = pd.Series(
            [
                "word word word word.",
                "This is \nthird line \nthird line",
            ],
        )

        expected = pd.Series([1, 4])
        actual = self.primitive().get_function()(x)
        pd.testing.assert_series_equal(actual, expected, check_names=False)

    def test_null(self):
        x = pd.Series([np.nan, pd.NA, None, "This is a test file."])

        actual = self.primitive().get_function()(x)
        expected = pd.Series([pd.NA, pd.NA, pd.NA, 5])
        pd.testing.assert_series_equal(actual, expected, check_names=False)

    def test_case_insensitive(self):
        x = pd.Series(["WORD word WORd WORd WOrD word"])

        actual = self.primitive(case_insensitive=True).get_function()(x)
        expected = pd.Series([1])
        pd.testing.assert_series_equal(actual, expected, check_names=False)

    def test_with_featuretools(self, es):
        transform, aggregation = find_applicable_primitives(self.primitive)
        primitive_instance = self.primitive()
        transform.append(primitive_instance)
        valid_dfs(es, aggregation, transform, self.primitive)



from datetime import datetime

import numpy as np
import pandas as pd
from woodwork.logical_types import (
    URL,
    Boolean,
    Categorical,
    CountryCode,
    Datetime,
    Double,
    EmailAddress,
    Filepath,
    Integer,
    IPAddress,
    LatLong,
    NaturalLanguage,
    Ordinal,
    PersonFullName,
    PhoneNumber,
    PostalCode,
    SubRegionCode,
)

from featuretools.entityset import EntitySet


def make_ecommerce_entityset(with_integer_time_index=False):
    """Makes a entityset with the following shape:

      R         Régions
     / \\       .
    S   C       Stores, Customers
        |       .
        S   P   Sessions, Products
         \\ /   .
          L     Log
    """
    dataframes = make_ecommerce_dataframes(
        with_integer_time_index=with_integer_time_index,
    )
    dataframe_names = dataframes.keys()
    es_id = "ecommerce"
    if with_integer_time_index:
        es_id += "_int_time_index"

    logical_types = make_logical_types(with_integer_time_index=with_integer_time_index)
    semantic_tags = make_semantic_tags()
    time_indexes = make_time_indexes(with_integer_time_index=with_integer_time_index)

    es = EntitySet(id=es_id)

    for df_name in dataframe_names:
        time_index = time_indexes.get(df_name, None)
        ti_name = None
        secondary = None
        if time_index is not None:
            ti_name = time_index["name"]
            secondary = time_index["secondary"]
        df = dataframes[df_name]
        es.add_dataframe(
            df,
            dataframe_name=df_name,
            index="id",
            logical_types=logical_types[df_name],
            semantic_tags=semantic_tags[df_name],
            time_index=ti_name,
            secondary_time_index=secondary,
        )

    es.normalize_dataframe(
        "customers",
        "cohorts",
        "cohort",
        additional_columns=["cohort_name"],
        make_time_index=True,
        new_dataframe_time_index="cohort_end",
    )

    es.add_relationships(
        [
            ("régions", "id", "customers", "région_id"),
            ("régions", "id", "stores", "région_id"),
            ("customers", "id", "sessions", "customer_id"),
            ("sessions", "id", "log", "session_id"),
            ("products", "id", "log", "product_id"),
        ],
    )

    return es


def make_ecommerce_dataframes(with_integer_time_index=False):
    region_df = pd.DataFrame(
        {"id": ["United States", "Mexico"], "language": ["en", "sp"]},
    )

    store_df = pd.DataFrame(
        {
            "id": range(6),
            "région_id": ["United States"] * 3 + ["Mexico"] * 2 + [np.nan],
            "num_square_feet": list(range(30000, 60000, 6000)) + [np.nan],
        },
    )

    product_df = pd.DataFrame(
        {
            "id": [
                "Haribo sugar-free gummy bears",
                "car",
                "toothpaste",
                "brown bag",
                "coke zero",
                "taco clock",
            ],
            "department": [
                "food",
                "electronics",
                "health",
                "food",
                "food",
                "electronics",
            ],
            "rating": [3.5, 4.0, 4.5, 1.5, 5.0, 5.0],
            "url": [
                "google.com",
                "https://www.featuretools.com/",
                "amazon.com",
                "www.featuretools.com",
                "bit.ly",
                "featuretools.com/demos/",
            ],
        },
    )
    customer_times = {
        "signup_date": [
            datetime(2011, 4, 8),
            datetime(2011, 4, 9),
            datetime(2011, 4, 6),
        ],
        # some point after signup date
        "upgrade_date": [
            datetime(2011, 4, 10),
            datetime(2011, 4, 11),
            datetime(2011, 4, 7),
        ],
        "cancel_date": [
            datetime(2011, 6, 8),
            datetime(2011, 10, 9),
            datetime(2012, 1, 6),
        ],
        "birthday": [datetime(1993, 3, 8), datetime(1926, 8, 2), datetime(1993, 4, 20)],
    }
    if with_integer_time_index:
        customer_times["signup_date"] = [6, 7, 4]
        customer_times["upgrade_date"] = [18, 26, 5]
        customer_times["cancel_date"] = [27, 28, 29]
        customer_times["birthday"] = [2, 1, 3]

    customer_df = pd.DataFrame(
        {
            "id": pd.Categorical([0, 1, 2]),
            "age": [33, 25, 56],
            "région_id": ["United States"] * 3,
            "cohort": [0, 1, 0],
            "cohort_name": ["Early Adopters", "Late Adopters", "Early Adopters"],
            "loves_ice_cream": [True, False, True],
            "favorite_quote": [
                "The proletariat have nothing to lose but their chains",
                "Capitalism deprives us all of self-determination",
                "All members of the working classes must seize the "
                "means of production.",
            ],
            "signup_date": customer_times["signup_date"],
            # some point after signup date
            "upgrade_date": customer_times["upgrade_date"],
            "cancel_date": customer_times["cancel_date"],
            "cancel_reason": ["reason_1", "reason_2", "reason_1"],
            "engagement_level": [1, 3, 2],
            "full_name": ["Mr. John Doe", "Doe, Mrs. Jane", "James Brown"],
            "email": ["john.smith@example.com", np.nan, "team@featuretools.com"],
            "phone_number": ["555-555-5555", "555-555-5555", "1-(555)-555-5555"],
            "birthday": customer_times["birthday"],
        },
    )

    ips = [
        "192.168.0.1",
        "2001:4860:4860::8888",
        "0.0.0.0",
        "192.168.1.1:2869",
        np.nan,
        np.nan,
    ]
    filepaths = [
        "/home/user/docs/Letter.txt",
        "./inthisdir",
        "C:\\user\\docs\\Letter.txt",
        "~/.rcinfo",
        "../../greatgrandparent",
        "data.json",
    ]

    session_df = pd.DataFrame(
        {
            "id": [0, 1, 2, 3, 4, 5],
            "customer_id": pd.Categorical([0, 0, 0, 1, 1, 2]),
            "device_type": [0, 1, 1, 0, 0, 1],
            "device_name": ["PC", "Mobile", "Mobile", "PC", "PC", "Mobile"],
            "ip": ips,
            "filepath": filepaths,
        },
    )

    times = list(
        [datetime(2011, 4, 9, 10, 30, i * 6) for i in range(5)]
        + [datetime(2011, 4, 9, 10, 31, i * 9) for i in range(4)]
        + [datetime(2011, 4, 9, 10, 40, 0)]
        + [datetime(2011, 4, 10, 10, 40, i) for i in range(2)]
        + [datetime(2011, 4, 10, 10, 41, i * 3) for i in range(3)]
        + [datetime(2011, 4, 10, 11, 10, i * 3) for i in range(2)],
    )
    if with_integer_time_index:
        times = list(range(8, 18)) + list(range(19, 26))

    values = list(
        [i * 5 for i in range(5)]
        + [i * 1 for i in range(4)]
        + [0]
        + [i * 5 for i in range(2)]
        + [i * 7 for i in range(3)]
        + [np.nan] * 2,
    )

    values_2 = list(
        [i * 2 for i in range(5)]
        + [i * 1 for i in range(4)]
        + [0]
        + [i * 2 for i in range(2)]
        + [i * 3 for i in range(3)]
        + [np.nan] * 2,
    )

    values_many_nans = list(
        [np.nan] * 5
        + [i * 1 for i in range(4)]
        + [0]
        + [np.nan] * 2
        + [i * 3 for i in range(3)]
        + [np.nan] * 2,
    )

    latlong = list([(values[i], values_2[i]) for i, _ in enumerate(values)])
    latlong2 = list([(values_2[i], -values[i]) for i, _ in enumerate(values)])
    zipcodes = list(
        ["02116"] * 5
        + ["02116-3899"] * 4
        + ["0"]
        + ["1234567890"] * 2
        + ["12345-6789"] * 2
        + [np.nan] * 3,
    )
    countrycodes = list(["US"] * 5 + ["AL"] * 4 + [np.nan] * 5 + ["ALB"] * 2 + ["USA"])
    subregioncodes = list(
        ["US-AZ"] * 5 + ["US-MT"] * 4 + [np.nan] * 3 + ["UG-219"] * 2 + ["ZM-06"] * 3,
    )
    log_df = pd.DataFrame(
        {
            "id": range(17),
            "session_id": [0] * 5 + [1] * 4 + [2] * 1 + [3] * 2 + [4] * 3 + [5] * 2,
            "product_id": ["coke zero"] * 3
            + ["car"] * 2
            + ["toothpaste"] * 3
            + ["brown bag"] * 2
            + ["Haribo sugar-free gummy bears"]
            + ["coke zero"] * 4
            + ["taco clock"] * 2,
            "datetime": times,
            "value": values,
            "value_2": values_2,
            "latlong": latlong,
            "latlong2": latlong2,
            "zipcode": zipcodes,
            "countrycode": countrycodes,
            "subregioncode": subregioncodes,
            "value_many_nans": values_many_nans,
            "priority_level": [0] * 2 + [1] * 5 + [0] * 6 + [2] * 2 + [1] * 2,
            "purchased": [True] * 11 + [False] * 4 + [True, False],
            "url": ["https://www.featuretools.com/"] * 2
            + ["amazon.com"] * 2
            + [
                "www.featuretools.com",
                "bit.ly",
                "featuretools.com/demos/",
                "www.google.co.in/" "http://lplay.google.co.in",
                " ",
                "invalid_url",
                "an",
                "microsoft.com/search/",
            ]
            + [np.nan] * 5,
            "email_address": ["john.smith@example.com", np.nan, "team@featuretools.com"]
            * 5
            + [" prefix@space.com", "suffix@space.com "],
            "comments": [coke_zero_review()]
            + ["I loved it"] * 2
            + car_reviews()
            + toothpaste_reviews()
            + brown_bag_reviews()
            + [gummy_review()]
            + ["I loved it"] * 4
            + taco_clock_reviews(),
        },
    )

    return {
        "régions": region_df,
        "stores": store_df,
        "products": product_df,
        "customers": customer_df,
        "sessions": session_df,
        "log": log_df,
    }


def make_semantic_tags():
    store_semantic_tags = {"région_id": "foreign_key"}

    customer_semantic_tags = {"région_id": "foreign_key", "birthday": "date_of_birth"}

    session_semantic_tags = {"customer_id": "foreign_key"}

    log_semantic_tags = {"session_id": "foreign_key"}

    return {
        "customers": customer_semantic_tags,
        "sessions": session_semantic_tags,
        "log": log_semantic_tags,
        "products": {},
        "stores": store_semantic_tags,
        "régions": {},
    }


def make_logical_types(with_integer_time_index=False):
    region_logical_types = {"id": Categorical, "language": Categorical}

    store_logical_types = {
        "id": Integer,
        "région_id": Categorical,
        "num_square_feet": Double,
    }

    product_logical_types = {
        "id": Categorical,
        "rating": Double,
        "department": Categorical,
        "url": URL,
    }

    customer_logical_types = {
        "id": Integer,
        "age": Integer,
        "région_id": Categorical,
        "loves_ice_cream": Boolean,
        "favorite_quote": NaturalLanguage,
        "signup_date": Datetime(datetime_format="%Y-%m-%d"),
        "upgrade_date": Datetime(datetime_format="%Y-%m-%d"),
        "cancel_date": Datetime(datetime_format="%Y-%m-%d"),
        "cancel_reason": Categorical,
        "engagement_level": Ordinal(order=[1, 2, 3]),
        "full_name": PersonFullName,
        "email": EmailAddress,
        "phone_number": PhoneNumber,
        "birthday": Datetime(datetime_format="%Y-%m-%d"),
        "cohort_name": Categorical,
        "cohort": Integer,
    }

    session_logical_types = {
        "id": Integer,
        "customer_id": Integer,
        "device_type": Categorical,
        "device_name": Categorical,
        "ip": IPAddress,
        "filepath": Filepath,
    }

    log_logical_types = {
        "id": Integer,
        "session_id": Integer,
        "product_id": Categorical,
        "datetime": Datetime(datetime_format="%Y-%m-%d"),
        "value": Double,
        "value_2": Double,
        "latlong": LatLong,
        "latlong2": LatLong,
        "zipcode": PostalCode,
        "countrycode": CountryCode,
        "subregioncode": SubRegionCode,
        "value_many_nans": Double,
        "priority_level": Ordinal(order=[0, 1, 2]),
        "purchased": Boolean,
        "url": URL,
        "email_address": EmailAddress,
        "comments": NaturalLanguage,
    }
    if with_integer_time_index:
        log_logical_types["datetime"] = Integer
        customer_logical_types["signup_date"] = Integer
        customer_logical_types["upgrade_date"] = Integer
        customer_logical_types["cancel_date"] = Integer
        customer_logical_types["birthday"] = Integer

    return {
        "customers": customer_logical_types,
        "sessions": session_logical_types,
        "log": log_logical_types,
        "products": product_logical_types,
        "stores": store_logical_types,
        "régions": region_logical_types,
    }


def make_time_indexes(with_integer_time_index=False):
    return {
        "customers": {
            "name": "signup_date",
            "secondary": {"cancel_date": ["cancel_reason"]},
        },
        "log": {"name": "datetime", "secondary": None},
    }


def coke_zero_review():
    return """
When it comes to Coca-Cola products, people tend to be die-hard fans. Many of us know someone who can't go a day without a Diet Coke (or two or three). And while Diet Coke has been a leading sugar-free soft drink since it was first released in 1982, it came to light that young adult males shied away from this beverage — identifying diet cola as a woman's drink. The company's answer to that predicament came in 2005 - in the form of a shiny black can - with the release of Coca-Cola Zero.

While Diet Coke was created with its own flavor profile and not as a sugar-free version of the original, Coca-Cola Zero aims to taste just like the "real Coke flavor." Despite their polar opposite advertising campaigns, the contents and nutritional information of the two sugar-free colas is nearly identical. With that information in hand we at HuffPost Taste needed to know: Which of these two artificially-sweetened Coca-Cola beverages actually tastes better? And can you even tell the difference between them?

Before we get to the results of our taste test, here are the facts:


Diet Coke

Motto: Always Great Tast
Nutritional Information: Many say that a can of Diet Coke actually contains somewhere between 1-4 calories, but if a serving size contains fewer than 5 calories a company is not obligated to note it in its nutritional information. Diet Coke's nutritional information reads 0 Calories, 0g Fat, 40mg Sodium, 0g Total Carbs, 0g Protein.

Ingredients: Carbonated water, caramel color, aspartame, phosphoric acid, potassium benzonate, natural flavors, citric acid, caffeine.

Artificial sweetener: Aspartame


Coca-Cola Zero
Motto: Real Coca-Cola Taste AND Zero Calories

Nutritional Information: While the label clearly advertises this beverage as a zero calorie cola, we are not entirely certain that its minimal calorie content is simply not required to be noted in the nutritional information. Coca-Cola Zero's nutritional information reads 0 Calories, 0g Fat, 40mg Sodium, 0g Total Carbs, 0g Protein.

Artificial sweetener: Aspartame and acesulfame potassium

Ingredients: Carbonated water, caramel color, phosphoric acid, aspartame, potassium benzonate, natural flavors, potassium citrate, acesulfame potassium, caffeine.

The Verdict:
Twenty-four editors blind-tasted the two cokes, side by side, and...

54 percent of our tasters were able to distinguish Diet Coke from Coca-Cola Zero
50 percent of our tasters preferred Diet Coke to Coca-Cola Zero, and vice versa
Here’s what our tasters thought of the two sugar-free soft drinks:

Diet Coke: "Tastes fake right away." "Much fresher brighter, crisper." "Has the wonderful flavors of Diet Coke’s artificial sweeteners."

Coca-Cola Zero: "Has more of a sharply sweet aftertaste I associate with diet sodas." "Tastes more like regular coke, less like fake sweetener." "Has an odd taste." "Tastes more like regular." "Very sweet."

Overall comments: "That was a lot more difficult than I though it would be." "Both equally palatable." A few people said Diet Coke tasted much better ... unbeknownst to them, they were actually referring to Coca-Cola Zero.

IN SUMMARY: It is a real toss up. There is not one artificially-sweetened Coca-Cola beverage that outshines the other. So how do people choose between one or the other? It is either a matter of personal taste, or maybe the marketing campaigns will influence their choice.
"""


def gummy_review():
    return """
The place: BMO Harris Bradley Center
The event: Bucks VS Spurs
The snack: Satan's Diarrhea Hate Bears made by Haribo

I recently took my 4 year old son to his first NBA game. He was very excited to go to the game, and I was excited because we had fantastic seats. Row C center court to be exact. I've never sat that close before. I've never had to go DOWN stairs to get to my seats. 24 stairs to get to my seats to be exact.

His favorite candy is Skittles. Mine are anything gummy. I snuck in a bag of skittles for my son, and grabbed a handful of gummy bears for myself, to be later known as Satan's Diarrhea Hate Bears, that I received for Christmas in bulk from my parents, and put them in a zip lock bag.

After the excitement of the 1st quarter has ended I take my son out to get him a bottled water and myself a beer. We return to our seats to enjoy our candy and drinks.

..............fast forward until 1 minute before half time...........

I have begun to sweat a sweat that is only meant for a man on mile 19 of a marathon. I have kicked out my legs out so straight that I am violently pushing the gentleman wearing a suit seat in front of me forward. He is not happy, I do not care. My hands are on the side of my seat not unlike that of a gymnast on a pommel horse, lifting me off my chair. My son is oblivious to what is happening next to him, after all, there is a mascot running around somewhere and he is eating candy.

I realize that at some point in the very near to immediate future I am going to have to allow this lava from Satan to forcefully expel itself from my innards. I also realize that I have to walk up 24 stairs just to get to level ground in hopes to make it to the bathroom. I’ll just have to sit here stiff as a board for a few moments waiting for the pain to subside. About 30 seconds later there is a slight calm in the storm of the violent hurricane that is going on in my lower intestine. I muster the courage to gently relax every muscle in my lower half and stand up. My son stands up next to me and we start to ascend up the stairs. I take a very careful and calculated step up the first stair. Then a very loud horn sounds. Halftime. Great. It’s going to be crowded. The horn also seems to have awaken the Satan's Diarrhea Hate Bears that are having a mosh pit in my stomach. It literally felt like an avalanche went down my stomach and I again have to tighten every muscle and stand straight up and focus all my energy on my poor sphincter to tighten up and perform like it has never performed before. Taking another step would be the worst idea possible, the flood gates would open. Don’t worry, Daddy has a plan. I some how mumble the question, “want to play a game?” to my son, he of course says “yes”. My idea is to hop on both feet allllll the way up the stairs, using the center railing to propel me up each stair. My son is always up for a good hopping game, so he complies and joins in on the “fun”. Some old lady 4 steps up thinks its cute that we are doing this, obviously she wasn’t looking at the panic on my face. 3 rows behind her a man about the same age as me, who must have had similar situations, notices the fear/panic/desperation on my face understands the danger that I along with my pants and anyone within a 5 yard radius spray zone are in. He just mouths the words “good luck man” to me and I press on. Half way up and there is no leakage, but my legs are getting tired and my sphincter has never endured this amount of pressure for this long of time. 16 steps/hops later…….4 steps to go…….My son trips and falls on the stairs, I have two options: keep going knowing he will catch up or bend down to pick him up relieving my sphincter of all the pressure and commotion while ruining the day of roughly the 50 people that are now watching a grown man hop up stairs while sweating profusely next to a 4 year old boy.

Luckily he gets right back up and we make it to the top of the stairs. Good, the hard part was over. Or so I thought. I managed to waddle like a penguin, or someone who is about to poop their pants in 2.5 seconds, to the men's room only to find that every stall is being used. EVERY STALL. It's halftime, of course everyone has to poop at that moment. I don't know if I can wait any longer, do I go ahead and fulfil the dream of every high school boy and poop in the urinal? What kind of an example would that set for my son? On the other hand, what kind of an example would it be for his father to fill his pants with a substance that probably will be unrecognizable to man. Suddenly a stall door opens, and I think I manage to actually levitate over to the stall. I my son follows me in, luckily it was the handicap stall so there was room for him to be out of the way. I get my pants off and start to sit. I know what taking a giant poo feels like. I also know what vomiting feels like. I can now successfully say that I know what it is like to vomit out my butt. I wasn't pooping, those Satan's Diarrhea Hate Bears did something to my insides that made my sphincter vomit our the madness.

I am now conscious of my surroundings. Other than the war that the bottom half of my body is currently having with this porcelain chair, it is quiet as a pin drop in the bathroom. The other men in there can sense that something isn't right, no one has heard anyone ever poop vomit before.

I can sense that the worst part is over. But its not stopping, nor can I physically stop it at this point, I am leaking..it's horrible. I call out "does anyone have a diaper?" hoping that some gentleman was changing a baby. Nothing. No one said a word. I know people are in there, I can see the toes of shoes pointed in my direction under the stall.. "DOES ANYONE HAVE A DIAPER!?!" I am screaming, my son is now crying, he thinks he is witnessing the death of his father. I can't even assure him that I will make it.

Not a word was said, but a diaper was thrown over the stall. I catch it, line my underwear with it, put my pants back on, and walk out of that bathroom like a champ. We go straight to our seats, grab out coats and go home. As we are walking out, the gentleman that wished me good luck earlier simply put his fist out, and I happily bumped it.

My son asks me, "Daddy, why are we leaving early?"
"Well son, I need to change my diaper"
"""


def taco_clock_reviews():
    return [
        """
This timer does what it is supposed to do. Setup is elementary. Replacing the old one (after 12 years) was relatively easy. It has performed flawlessly since. I'm delighted I could find an esoteric product like this at Amazon. Their service, and the customer reviews, are just excellent.
""",
        """
Funny, cute clock. A little spendy for how light the clock is, but its hard to find a taco clock.
""",
    ]


def brown_bag_reviews():
    return [
        """
These bags looked exactly like I'd hoped, however, the handles broke off of almost every single bag as soon as items were placed in them! I used these as gift bags for out-of-town guests at my wedding, so imagine my embarassment as the handles broke off as I was handing them out. I would not recommend purchaing these bags unless you plan to fill them with nothing but paper! Anything heavier will cause the handles to snap right off.
""",
        """
I purchased these in August 2014 from Big Blue Supplies. I have no problem with the seller, these arrived new condition, fine shape.

I do have a slight problem with the bags. In case someone might want to know, the handles on these bags are set inside against the top. Then a piece of Kraft type packing tape is placed over the handles to hold them in place. On some of the bags, the tape is already starting to peel off. I would be really hesitant about using these bags unless I reinforced the current tape with a different adhesive.

I will keep the bags, and make a tape of a holiday or decorative theme and place over in order to make certain the handles stay in place.

Also in case anybody is wondering, the label on the plastic packaging bag states these are from ORIENTAL TRADING COMPANY. On the bottom of each bag it is stamped MADE IN CHINA. Again, I will be placing a sticker over that.

Even the dollar store bags I normally purchase do not have that stamped on the bottom in such prominent lettering. I purchased these because they were plain and I wanted to decorate them.

I do not think I would purchase again for all the reasons stated above.

Another thing for those still wanting to purchase, the ones I received were: 12 3/4 inches high not including handle, 10 1/4 inches wide and a 5 1/4 inch depth.
""",
    ]


def car_reviews():
    return [
        """
The full-size pickup truck and the V-8 engine were supposed to be inseparable, like the internet and cat videos. You can’t have one without the other—or so we thought.

In America’s most popular vehicle, the Ford F-150, two turbocharged six-cylinder engines marketed under the EcoBoost name have dethroned the naturally aspirated V-8. Ford’s new 2.7-liter twin-turbo V-6 is the popular choice, while the 3.5-liter twin-turbo V-6 is the top performer. The larger six allows for greater hauling capacity, accelerates the truck more quickly, and swills less gas in EPA testing than the V-8 alternative. It’s enough to make even old-school truck buyers acknowledge that there actually is a replacement for displacement.

And yet a V-8 in a big pickup truck still feels so natural, so right. In the F-150, the Coyote 5.0-liter V-8 is tuned for torque more so than power, yet it still revs with an enthusiastic giddy-up that reminds us that this engine’s other job is powering the Mustang. The response follows the throttle pedal faithfully while the six-speed automatic clicks through gears smoothly and easily. Together they pull this 5220-pound F-150 to 60 mph in 6.3 seconds, which is 0.4 second quicker than the 5.3-liter Chevrolet Silverado with the six-speed automatic and 0.9 second quicker than the 5.3 Silverado with the new eight-speed auto. The 3.5-liter EcoBoost, though, can do the deed another half-second quicker, but its synthetic soundtrack doesn’t have the rich, multilayered tone of the V-8.

It wasn’t until we saddled our test truck with a 6400-pound trailer (well under its 9000-pound rating) that we fully understood the case for upgrading to the 3.5-liter EcoBoost. The twin-turbo engine offers an extra 2500 pounds of towing capability and handles lighter tasks with considerably less strain. The 5.0-liter truck needs more revs and a wider throttle opening to accelerate its load, so we were often coaxed into pressing the throttle to the floor for even modest acceleration. The torquier EcoBoost engine offers a heartier response at part throttle.

In real-world, non-towing situations, the twin-turbo 3.5-liter doesn’t deliver on its promise of increased fuel economy, with both the 5.0-liter V-8 and that V-6 returning 16 mpg in our hands. But given the 3.5-liter’s virtues, we can forgive it that trespass.

Trucks Are the New Luxury

Pickups once were working-class transportation. Today, they’re proxy luxury vehicles—or at least that’s how they’re priced. If you think our test truck’s $57,240 window sticker is steep, consider that our model, the Lariat, is merely a mid-spec trim. There are three additional grades—King Ranch, Platinum, and Limited—positioned and priced above it, plus the 3.5-liter EcoBoost that costs an extra $400 as well as a plethora of options to inflate the price past 60 grand. Squint and you can almost see the six-figure trucks of the future on the horizon.

For the most part, though, the equipment in this particular Lariat lives up to the price tag. The driver and passenger seats are heated and cooled, with 10-way power adjustability and supple leather. The technology includes blind-spot monitoring, navigation, and a 110-volt AC outlet. Nods to utility include spotlights built into the side mirrors and Ford’s Pro Trailer Backup Assist, which makes reversing with a trailer as easy as turning a tiny knob on the dashboard.

Middle-Child Syndrome

In the F-150, Ford has a trifecta of engines (the fourth, a naturally aspirated 3.5-liter V-6, is best left to the fleet operators). The 2.7-liter twin-turbo V-6 delivers remarkable performance at an affordable price. The 3.5-liter twin-turbo V-6 is the workhorse, with power, torque, and hauling capability to spare. Compared with those two logical options, the middle-child 5.0-liter V-8 is the right-brain choice. Its strongest selling points may be its silky power delivery and the familiar V-8 rumble. That’s a flimsy argument when it comes to rationalizing a $50,000-plus purchase, though, so perhaps it’s no surprise that today’s boosted six-cylinders are now the engines of choice in the F-150.
""",
        """
THE GOOD
The Tesla Model S 90D's electric drivetrain is substantially more efficient than any internal combustion engine, and gives the car smooth and quick acceleration. All-wheel drive comes courtesy of a smart dual motor system. The new Autopilot feature eases the stress of stop-and-go traffic and long road trips.

THE BAD
Even at Tesla's Supercharger stations, recharging the battery takes significantly longer than refilling an internal combustion engine car's gas tank, limiting where you can drive. Tesla hasn't improved its infotainment system much from the Model S' launch.

THE BOTTOM LINE
Among the different flavors of Tesla Model S, the 90D is the one to get, exhibiting the best range and all-wheel drive, while offering an uncomplicated, next-generation driving experience that shows very well against equally priced competitors.


REVIEW  SPECIFICATIONS  PHOTOS
Roadshow Automobiles Tesla 2016 Tesla Model S
Having tested driver assistance systems in many cars, and even ridden in fully self-driving cars, I should have been ready for Tesla's new Autopilot feature. But engaging it while cruising the freeway in the Model S 90D, I kept my foot hovering over the brake.

My trepidation didn't come so much from the adaptive cruise control, which kept the Model S following traffic ahead at a set distance, but from the self-steering, this part of Autopilot managing to keep the Model S well-centered in its lane with no help from me. Over many miles, I built up more trust in the system, letting the car do the steering in situations from bumper-to-bumper traffic and a winding road through the hills.

2016 Tesla Model S 90DEnlarge Image
Although the middle of the Model S range, the 90D offers the best range and a wealth of useful tech, such as Autopilot self-driving.
Wayne Cunningham/Roadshow
Tesla added Autopilot to its Model S line as an option last year, along with all-wheel-drive. More recently, the high-tech automaker improved its batteries, upgrading its cars from their former 65 and 85 kilowatt-hour capacity to 70 and 90 kilowatt-hour. The example I drove, the 90D, represents all these advances.

More importantly, the 90D is the current range-leader among the Model S line, boasting 288 miles on a full battery charge.

The Model S' improvements fall outside of typical automotive industry product cycles, fulfilling Tesla's promise of acting more like a technology company, constantly building and deploying new features. Tesla accomplishes that goal partially through over-the-air software updates, improving existing cars, but the 90D presents significant hardware updates over the original Model S launched four years ago.

Sit and go
Of course, this Model S exhibited the ease of use of the original. Walking up to the car with the key fob in my pocket, it automatically unlocked. When I got in the car, it powered up without me having to push a start button, so I only needed to put it in drive to get on the road.

Likewise, the design hasn't changed, its sleek, hatchback four-door body offering excellent cargo room, both front and back, and seating space. The cabin feels less cramped than most cars due to the lack of a transmission tunnel and a dashboard bare of buttons or dials.

2016 Tesla Model S 90DEnlarge Image
The flat floor in the Model S' cabin makes for enhanced passenger room.
Wayne Cunningham/Roadshow
The big, 17-inch touchscreen in the center of the dashboard shows navigation, stereo, phone, energy consumption and car settings. I easily went from full-screen to a split-screen view, the windows showing each appearing instantly. A built-in 4G/LTE data connection powers Google maps and Internet-based audio. The LCD instrument panel in front of me showed my speed, energy usage, remaining range, and intelligently swapped audio information for turn-by-turn directions when started navigation.

The instrument panel actually made the experience of driving under Autopilot more comfortable, reassuring me with graphics that showed when the Model S' sensors were detecting the lane lines and the traffic around me. Impressively, the sensors could differentiate, as shown on the screen's graphics, a passenger car from a big truck.

At speed on the freeway, Autopilot smoothly maintained the car's position in its lane, and when I took my hands off the wheel for too long, it flashed a warning on the instrument panel. In stop-and-go traffic approaching a toll booth, the car did an even better job of self-driving, recognizing traffic around it and maintaining appropriate distances.

Handling surprise
Taking over the driving myself, the ride quality proved as comfortable as any sport-luxury car, as this Model S had its optional air suspension. The electric power steering is well-tuned, turning the wheels with a quiet, natural feel and good heft.

Audi S7 vs Tesla Model S
Shootout: Audi S7 vs. Tesla Model S
Wayne Cunningham/Roadshow
The biggest surprise came when I spent the day doing laps at the Thunderhill Raceway, negotiating a series of tight, technical turns in competition with an Audi S7. I expected the Model S to get out-of-shape in the turns, but instead it proved steady and solid. The Model S' 4,647-pound curb weight made it less than ideal for a track test, but much of that weight is in the battery pack, mounted low in the chassis. That low center of gravity helped limit body roll, ensuring good grip from all four tires. In the turns, the Model S felt nicely balanced, although not entirely nimble.

Helping its grip was its native all-wheel drive, gained from having motors driving each set of wheels. The combined output of the motors comes to 417 horsepower and 485 pound-feet of torque, those numbers expressed in 0-to-60 mph times of well under 5 seconds. That thrust made for fast runs down the race track's straightaways, or simply giving me the ability to take advantage of gaps in traffic on public roads.

288 miles is more than enough for most people's daily driving needs, and if you plug in every night, you will wake up to a fully charged car every morning. The Model S makes for a far different experience than driving an internal combustion car, where you need to go to a gas station to refuel. However, longer trips in the Model S require some planning, such as scheduling stops at Tesla's free Supercharger stations.


Charging times are much lengthier than refilling a tank with gasoline. From a Level 2, 240-volt station, you get 29 miles added every hour. Tesla's Supercharger, a Level 3 charger, takes 75 minutes to fully recharge the Model S 90D's battery.

2016 Tesla Model S 90DEnlarge Image
Despite its high initial price, the Model S 90D costs less to run on a daily basis than a combustion engine car.
Wayne Cunningham/Roadshow

Low maintenance
The 2016 Tesla Model S 90D adds features to keep it competitive against the internal combustion cars in its sport luxury set. More importantly, it remains very easy to live with. In fact, the electric drivetrain should mean greatly decreased maintenance, as there are fewer moving parts. The EPA estimates that annual electricity costs for the Model S 90D should run $650, much less than buying gasoline for an equivalent internal combustion car.

Lengthy charging times mean longer trips are either out of the question or require more planning than with an internal combustion car. And while the infotainment system responds quickly to touch inputs and offers useful screens, it hasn't changed much in four years. Most notably, Tesla hasn't added any music apps beyond the ones it launched with. Along with new, useful apps, it would be nice to have some themes or other aesthetic changes to the infotainment interface.

The Model S 90D's base price of $88,000 puts it out of reach of the average buyer, and the model I drove was optioned up to around $95,000. Against its Audi, BMW and Mercedes-Benz competition, however, it makes a compelling argument, especially for its uncomplicated nature.
""",
    ]


def toothpaste_reviews():
    return [
        """
Toothpaste can do more harm than good

The next time a patient innocently asks me, “What’s the best toothpaste to use?” I’m going to unleash a whole Chunky Soup can of “You Want The Truth? You CAN’T HANDLE THE TRUTH!!!” Gosh, that’s such an overused movie quote. Sorry about that, but still.

If you’re a dental professional, isn’t this the most annoying question you get, day after day? Do you even care which toothpaste your patients use?

No. You don’t. Asking a dentist what toothpaste to use is like asking your physician which bar of soap or body scrub you should use to clean your skin. Your dentist and dental hygienist have never seen a tube of toothpaste that singlehandedly improves the health of all patients in their practice, and the reason is simple:

Toothpaste is a cosmetic.

We brush our teeth so that out mouths no longer taste like… mouth. Mouth tastes gross, right? It tastes like putrefied skin. It tastes like tongue cheese. It tastes like Cream of Barf.

On the other hand, toothpaste has been exquisitely designed to bring you a brisk rush of York Peppermint Patty, or Triple Cinnamon Heaven, or whatever flavor that drives those tubes off of the shelves in the confusing dental aisle of your local supermarket or drugstore.


Toothpaste definitely tastes better than Cream of Barf. And that’s why you use it. Not because it’s good for you. You use toothpaste because it tastes good, and because it makes you accept your mouth as part of your face again.

From a marketing perspective, all of the other things that are in your toothpaste are in there to give it additional perceived value. So let’s deconstruct these ingredients, shall we?


1. Fluoride.

This was probably the first additive to toothpaste that brought it under the jurisdiction of the Food & Drug Administration and made toothpaste part drug, part cosmetic. Over time, a fluoride toothpaste can improve the strength of teeth, but the fluoride itself does nothing to make teeth cleaner. Some people are scared of fluoride so they don’t use it. Their choice. Professionally speaking, I know that the benefits of a fluoride additive far outweigh the risks.

2. Foam.

Sodium Lauryl Sulfate is soap. Soap has a creamy, thick texture that American tongues especially like and equate to the feeling of cleanliness. There’s not enough surfactant, though, in toothpaste foam to break up the goo that grows on your teeth. If these bubbles scrubbed, you’d better believe that they would also scrub your delicate gum tissues into a bloody pulp.

3. Abrasive particles.

Most toothpastes use hydrated silica as the grit that polishes teeth. You’re probably most familiar with it as the clear beady stuff in the “Do Not Eat” packets. Depending on the size and shape of the particles, silica is the whitening ingredient in most whitening toothpastes. But whitening toothpaste cannot get your teeth any whiter than a professional dental cleaning, because it only cleans the surface. Two weeks to a whiter smile? How about 30 minutes with your hygienist? It’s much more efficient and less harsh.

4. Desensitizers.

Teeth that are sensitive to hot, cold, sweets, or a combination can benefit from the addition of potassium nitrate or stannous fluoride to a toothpaste. This is more of a palliative treatment, when the pain is the problem. Good old Time will usually make teeth feel better, too, unless the pain is coming from a cavity. Yeah, I’m talking to you, the person who is trying to heal the hole in their tooth with Sensodyne.

5. Tartar control.

It burns! It burns! If your toothpaste has a particular biting flavor, it might contain tetrasodium pyrophosphate, an ingredient that is supposed to keep calcium phosphate salts (tartar, or calculus) from fossilizing on the back of your lower front teeth. A little tartar on your teeth doesn’t harm you unless it gets really thick and you can no longer keep it clean. One problem with tartar control toothpastes is that in order for the active ingredient to work, it has to be dissolved in a stronger detergent than usual, which can affect people that are sensitive to a high pH.

6. Triclosan.

This antimicrobial is supposed to reduce infections between the gum and tooth. However, if you just keep the germs off of your teeth in the first place it’s pretty much a waste of an extra ingredient. Its safety has been questioned but, like fluoride, the bulk of the scientific research easily demonstrates that the addition of triclosan in toothpaste does much more good than harm.

Why toothpaste can be bad for you.

Let’s just say it’s not the toothpaste’s fault. It’s yours. The toothpaste is just the co-dependent enabler. You’re the one with the problem.

Remember, toothpaste is a cosmetic, first and foremost. It doesn’t clean your teeth by itself. Just in case you think I’m making this up I’ve included clinical studies in the references at the end of this article that show how ineffective toothpaste really is.

peasized

• You’re using too much.

Don’t be so suggestible! Toothpaste ads show you how to use up the tube more quickly. Just use 1/3 as much, the size of a pea. It will still taste good, I promise! And too much foam can make you lose track of where your teeth actually are located.

• You’re not taking enough time.

At least two minutes. Any less and you’re missing spots. Just ’cause it tastes better doesn’t mean you did a good job.

• You’re not paying attention.

I’ve seen people brush the same four spots for two minutes and miss the other 60% of their mouth.brushguide The toothbrush needs to touch every crevice of every tooth, not just where it lands when you go into autopilot and start thinking about what you’re going to wear that day. It’s the toothbrush friction that cleans your teeth, not the cleaning product. Plaque is a growth, like the pink or grey mildew that grows around the edges of your shower. You’ve gotta rub it off to get it off. No tooth cleaning liquid, paste, creme, gel, or powder is going to make as much of a difference as your attention to detail will.

The solution.

Use what you like. It’s that simple. If it tastes good and feels clean to you, you’ll use it more often, brush longer, feel better, be healthier.

You can use baking soda, or coconut oil, or your favorite toothpaste, or even just plain water. The key is to have a good technique and to brush often. A music video makes this demonstration a little more fun than your usual lecture at the dental office, although, in my opinion you really still need to feel what it is like to MASH THE BRISTLES OF A SOFT TOOTHBRUSH INTO YOUR GUMS:





A little more serious video from my pal Dr. Mark Burhenne where he demonstrates how to be careful with your toothbrush bristles:


Final word.

♬ It’s all about that Bass, ’bout that Bass, no bubbles. ♬ Heh, dentistry in-joke there.

Seriously, though, the bottom line is that your paste will mask brushing technique issues, so don’t put so much faith in the power of toothpaste.

Also you may have heard that some toothpastes contain decorative plastic that can get swallowed. Yeah, that was a DentalBuzz report I wrote that went viral earlier this year. And while I can’t claim total victory on that front, at least the company in question has promised that the plastic will no longer be added to their toothpaste lines very soon due to the overwhelming amount of letters, emails, and phone calls that they received as a result of people reading that article and making a difference.

But now I’m tired of talking about toothpaste.

Next topic?

I’m bringing pyorrhea back.
    """,
        """
I’ve been a user of Colgate Total Whitening Toothpaste for many years because I’ve always tried to maintain a healthy smile (I’m a receptionist so I need a white smile). But because I drink coffee at least twice a day (sometimes more!) and a lot of herbal teas, I’ve found that using just this toothpaste alone doesn’t really get my teeth white...

The best way to get white teeth is to really try some professional products specifically for tooth whitening. I’ve tried a few products, like Crest White Strips and found that the strips are really not as good as the trays. Although the Crest White Strips are easy to use, they really DO NOT cover your teeth perfectly like some other professional dental whitening kits. This Product did cover my teeth well however because of their custom heat trays, and whitening my teeth A LOT. I would say if you really want white teeth, use the Colgate Toothpaste and least 2 times a day, along side a professional Gel product like Shine Whitening.
    """,
        """
The first feature is the price, and it is right.

Next, I consider whether it will be neat to use. It is. Sometimes when I buy those new hard plastic containers, they actually get messy. Also I cannot get all the toothpaste out. It is easy to get the paste out of Colgate Total Whitening Paste without spraying it all over the cabinet.

If it does not taste good, I won't use it. Some toothpaste burns my mouth so bad that brushing my teeth is a painful experience. This one doesn't burn. It tastes simply the way toothpaste is supposed to taste.

Whitening is important. This one is supposed ot whiten. After spending money to whiten my teeth, I need a product to help ward off the bad effects of coffee and tea.

Avoiding all kinds of oral pathology is a major consideration. This toothpaste claims that it can help fight cavities, gingivitis, plaque, tartar, and bad breath.

I hope this product stays on the market a long time and does not change.
    """,
    ]



import pandas as pd
from dateutil.relativedelta import relativedelta


class Timedelta(object):
    """Represents differences in time.

    Timedeltas can be defined in multiple units. Supported units:

    - "ms" : milliseconds
    - "s" : seconds
    - "h" : hours
    - "m" : minutes
    - "d" : days
    - "o"/"observations" : number of individual events
    - "mo" : months
    - "Y" : years

    Timedeltas can also be defined in terms of observations. In this case, the
    Timedelta represents the period spanned by `value`.

    For observation timedeltas:
    >>> three_observations_log = Timedelta(3, "observations")
    >>> three_observations_log.get_name()
    '3 Observations'
    """

    _Observations = "o"

    # units for absolute times
    _absolute_units = ["ms", "s", "h", "m", "d", "w"]
    _relative_units = ["mo", "Y"]

    _readable_units = {
        "ms": "Milliseconds",
        "s": "Seconds",
        "h": "Hours",
        "m": "Minutes",
        "d": "Days",
        "o": "Observations",
        "w": "Weeks",
        "Y": "Years",
        "mo": "Months",
    }

    _readable_to_unit = {v.lower(): k for k, v in _readable_units.items()}

    def __init__(self, value, unit=None, delta_obj=None):
        """
        Args:
            value (float, str, dict) : Value of timedelta, string providing
                both unit and value, or a dictionary of units and times.
            unit (str) : Unit of time delta.
            delta_obj (pd.Timedelta or pd.DateOffset) : A time object used
                internally to do time operations. If None is provided, one will
                be created using the provided value and unit.
        """
        self.check_value(value, unit)
        self.times = self.fix_units()

        if delta_obj is not None:
            self.delta_obj = delta_obj
        else:
            self.delta_obj = self.get_unit_type()

    @classmethod
    def from_dictionary(cls, dictionary):
        dict_units = dictionary["unit"]
        dict_values = dictionary["value"]
        if isinstance(dict_units, str) and isinstance(dict_values, (int, float)):
            return cls({dict_units: dict_values})
        else:
            all_units = dict()
            for i in range(len(dict_units)):
                all_units[dict_units[i]] = dict_values[i]
            return cls(all_units)

    @classmethod
    def make_singular(cls, s):
        if len(s) > 1 and s.endswith("s"):
            return s[:-1]
        return s

    @classmethod
    def _check_unit_plural(cls, s):
        if len(s) > 2 and not s.endswith("s"):
            return (s + "s").lower()
        elif len(s) > 1:
            return s.lower()
        return s

    def get_value(self, unit=None):
        if unit is not None:
            return self.times[unit]
        elif len(self.times.values()) == 1:
            return list(self.times.values())[0]
        else:
            return self.times

    def get_units(self):
        return list(self.times.keys())

    def get_unit_type(self):
        all_units = self.get_units()
        if self._Observations in all_units:
            return None
        elif self.is_absolute() and self.has_multiple_units() is False:
            return pd.Timedelta(self.times[all_units[0]], all_units[0])
        else:
            readable_times = self.lower_readable_times()
            return relativedelta(**readable_times)

    def check_value(self, value, unit):
        if isinstance(value, str):
            from featuretools.utils.wrangle import _check_timedelta

            td = _check_timedelta(value)
            self.times = td.times
        elif isinstance(value, dict):
            self.times = value
        else:
            self.times = {unit: value}

    def fix_units(self):
        fixed_units = dict()
        for unit, value in self.times.items():
            unit = self._check_unit_plural(unit)
            if unit in self._readable_to_unit:
                unit = self._readable_to_unit[unit]
            fixed_units[unit] = value
        return fixed_units

    def lower_readable_times(self):
        readable_times = dict()
        for unit, value in self.times.items():
            readable_unit = self._readable_units[unit].lower()
            readable_times[readable_unit] = value
        return readable_times

    def get_name(self):
        all_units = self.get_units()
        if self.has_multiple_units() is False:
            return "{} {}".format(
                self.times[all_units[0]],
                self._readable_units[all_units[0]],
            )
        final_str = ""
        for unit, value in self.times.items():
            if value == 1:
                unit = self.make_singular(unit)
            final_str += "{} {} ".format(value, self._readable_units[unit])
        return final_str[:-1]

    def get_arguments(self):
        units = list()
        values = list()
        for unit, value in self.times.items():
            units.append(unit)
            values.append(value)
        if len(units) == 1:
            return {"unit": units[0], "value": values[0]}
        else:
            return {"unit": units, "value": values}

    def is_absolute(self):
        for unit in self.get_units():
            if unit not in self._absolute_units:
                return False
        return True

    def has_no_observations(self):
        for unit in self.get_units():
            if unit in self._Observations:
                return False
        return True

    def has_multiple_units(self):
        if len(self.get_units()) > 1:
            return True
        else:
            return False

    def __eq__(self, other):
        if not isinstance(other, Timedelta):
            return False

        return self.times == other.times

    def __neg__(self):
        """Negate the timedelta"""
        new_times = dict()
        for unit, value in self.times.items():
            new_times[unit] = -value
        if self.delta_obj is not None:
            return Timedelta(new_times, delta_obj=-self.delta_obj)
        else:
            return Timedelta(new_times)

    def __radd__(self, time):
        """Add the Timedelta to a timestamp value"""
        if self._Observations not in self.get_units():
            return time + self.delta_obj
        else:
            raise Exception("Invalid unit")

    def __rsub__(self, time):
        """Subtract the Timedelta from a timestamp value"""
        if self._Observations not in self.get_units():
            return time - self.delta_obj
        else:
            raise Exception("Invalid unit")



from woodwork.column_schema import ColumnSchema
from woodwork.logical_types import Datetime, Integer

from featuretools.primitives.base import AggregationPrimitive


class NUniqueDays(AggregationPrimitive):
    """Determines the number of unique days.

    Description:
        Given a list of datetimes, return the number of unique days.
        The same day in two different years is treated as different. So
        Feb 21, 2017 is different than Feb 21, 2019, even though they are
        both the 21st of February.

    Examples:
        >>> from datetime import datetime
        >>> n_unique_days = NUniqueDays()
        >>> times = [datetime(2019, 2, 1),
        ...          datetime(2019, 2, 1),
        ...          datetime(2018, 2, 1),
        ...          datetime(2019, 1, 1)]
        >>> n_unique_days(times)
        3
    """

    name = "n_unique_days"
    input_types = [ColumnSchema(logical_type=Datetime)]
    return_type = ColumnSchema(logical_type=Integer, semantic_tags={"numeric"})
    stack_on_self = False
    default_value = 0

    def get_function(self):
        def n_unique_days(x):
            return x.dt.floor("D").nunique()

        return n_unique_days



from datetime import datetime

import numpy as np
import pandas as pd
from pytest import raises

from featuretools.primitives import IsFederalHoliday


def test_regular():
    primitive_instance = IsFederalHoliday()
    primitive_func = primitive_instance.get_function()
    case = pd.Series(
        [
            "2016-01-01",
            "2016-02-29",
            "2017-05-29",
            datetime(2019, 7, 4, 10, 0, 30),
        ],
    ).astype("datetime64[ns]")
    answer = pd.Series([True, False, True, True])
    given_answer = pd.Series(primitive_func(case))
    assert given_answer.equals(answer)


def test_nat():
    primitive_instance = IsFederalHoliday()
    primitive_func = primitive_instance.get_function()
    case = pd.Series(
        [
            "2019-10-14",
            "NaT",
            "2016-02-29",
            "NaT",
        ],
    ).astype("datetime64[ns]")
    answer = pd.Series([True, np.nan, False, np.nan])
    given_answer = pd.Series(primitive_func(case))
    assert given_answer.equals(answer)


def test_valid_country():
    primitive_instance = IsFederalHoliday(country="Canada")
    primitive_func = primitive_instance.get_function()
    case = pd.Series(
        [
            "2016-07-01",
            "2016-11-11",
            "2017-12-26",
            "2018-09-03",
        ],
    ).astype("datetime64[ns]")
    answer = pd.Series([True, False, True, True])
    given_answer = pd.Series(primitive_func(case))
    assert given_answer.equals(answer)


def test_invalid_country():
    error_text = "must be one of the available countries"
    with raises(ValueError, match=error_text):
        IsFederalHoliday(country="")


def test_multiple_countries():
    primitive_mexico = IsFederalHoliday(country="Mexico")
    primitive_func = primitive_mexico.get_function()
    case = pd.Series([datetime(2000, 9, 16), datetime(2005, 1, 1)])
    assert len(primitive_func(case)) > 1
    primitive_india = IsFederalHoliday(country="IND")
    primitive_func = primitive_mexico.get_function()
    case = pd.Series([datetime(2048, 1, 1), datetime(2048, 10, 2)])
    primitive_func = primitive_india.get_function()
    assert len(primitive_func(case)) > 1
    primitive_uk = IsFederalHoliday(country="UK")
    primitive_func = primitive_uk.get_function()
    case = pd.Series([datetime(2048, 3, 17), datetime(2048, 4, 6)])
    assert len(primitive_func(case)) > 1
    countries = [
        "Argentina",
        "AU",
        "Austria",
        "BY",
        "Belgium",
        "Brazil",
        "Canada",
        "Colombia",
        "Croatia",
        "England",
        "Finland",
        "FRA",
        "Germany",
        "Germany",
        "Italy",
        "NewZealand",
        "PortugalExt",
        "PTE",
        "Spain",
        "ES",
        "Switzerland",
        "UnitedStates",
        "US",
        "UK",
        "UA",
        "CH",
        "SE",
        "ZA",
    ]
    for x in countries:
        IsFederalHoliday(country=x)



from featuretools.utils.description_utils import convert_to_nth


def test_first():
    assert convert_to_nth(1) == "1st"
    assert convert_to_nth(21) == "21st"
    assert convert_to_nth(131) == "131st"


def test_second():
    assert convert_to_nth(2) == "2nd"
    assert convert_to_nth(22) == "22nd"
    assert convert_to_nth(232) == "232nd"


def test_third():
    assert convert_to_nth(3) == "3rd"
    assert convert_to_nth(23) == "23rd"
    assert convert_to_nth(133) == "133rd"


def test_nth():
    assert convert_to_nth(4) == "4th"
    assert convert_to_nth(11) == "11th"
    assert convert_to_nth(12) == "12th"
    assert convert_to_nth(13) == "13th"
    assert convert_to_nth(111) == "111th"
    assert convert_to_nth(112) == "112th"
    assert convert_to_nth(113) == "113th"



import os
from inspect import signature

import numpy as np
import pandas as pd

from featuretools import config
from featuretools.utils.description_utils import convert_to_nth
from featuretools.utils.gen_utils import Library


class PrimitiveBase(object):
    """Base class for all primitives."""

    #: (str): Name of the primitive
    name = None
    #: (list): woodwork.ColumnSchema types of inputs
    input_types = None
    #: (woodwork.ColumnSchema): ColumnSchema type of return
    return_type = None
    #: Default value this feature returns if no data found. Defaults to np.nan
    default_value = np.nan
    #: (bool): True if feature needs to know what the current calculation time
    # is (provided to computational backend as "time_last")
    uses_calc_time = False
    #: (int): Maximum number of features in the largest chain proceeding
    # downward from this feature's base features.
    max_stack_depth = None
    #: (int): Number of columns in feature matrix associated with this feature
    number_output_features = 1
    # whitelist of primitives can have this primitive in input_types
    base_of = None
    # blacklist of primitives can have this primitive in input_types
    base_of_exclude = None
    # whitelist of primitives that can be in input_types
    stack_on = None
    # blacklist of primitives that can be in signature
    stack_on_exclude = None
    # determines if primitive can be in input_types for self
    stack_on_self = True
    # (bool) If True will only make one feature per unique set of base features
    commutative = False
    #: (list): Additional compatible libraries
    compatibility = [Library.PANDAS]
    #: (str, list[str]): description template of the primitive. Input column
    # descriptions are passed as positional arguments to the template. Slice
    # number (if present) in "nth" form is passed to the template via the
    # `nth_slice` keyword argument. Multi-output primitives can use a list to
    # differentiate between the base description and a slice description.
    description_template = None
    series_library = Library.PANDAS

    def __init__(self):
        pass

    def __call__(self, *args, **kwargs):
        series_args = [pd.Series(arg) for arg in args]
        try:
            return self._method(*series_args, **kwargs)
        except AttributeError:
            self._method = self.get_function()
            return self._method(*series_args, **kwargs)

    def __lt__(self, other):
        return (self.name + self.get_args_string()) < (
            other.name + other.get_args_string()
        )

    def generate_name(self):
        raise NotImplementedError("Subclass must implement")

    def generate_names(self):
        raise NotImplementedError("Subclass must implement")

    def get_function(self):
        raise NotImplementedError("Subclass must implement")

    def get_filepath(self, filename):
        return os.path.join(config.get("primitive_data_folder"), filename)

    def get_args_string(self):
        strings = []
        for name, value in self.get_arguments():
            # format arg to string
            string = "{}={}".format(name, str(value))
            strings.append(string)

        if len(strings) == 0:
            return ""

        string = ", ".join(strings)
        string = ", " + string
        return string

    def get_arguments(self):
        values = []

        args = signature(self.__class__).parameters.items()
        for name, arg in args:
            # assert that arg is attribute of primitive
            error = '"{}" must be attribute of {}'
            assert hasattr(self, name), error.format(name, self.__class__.__name__)

            value = getattr(self, name)
            # check if args are the same type
            if isinstance(value, type(arg.default)):
                # skip if default value
                if arg.default == value:
                    continue

            values.append((name, value))

        return values

    def get_description(
        self,
        input_column_descriptions,
        slice_num=None,
        template_override=None,
    ):
        template = template_override or self.description_template
        if template:
            if isinstance(template, list):
                if slice_num is not None:
                    slice_index = slice_num + 1
                    if slice_index < len(template):
                        return template[slice_index].format(
                            *input_column_descriptions,
                            nth_slice=convert_to_nth(slice_index),
                        )
                    else:
                        if len(template) > 2:
                            raise IndexError("Slice out of range of template")
                        return template[1].format(
                            *input_column_descriptions,
                            nth_slice=convert_to_nth(slice_index),
                        )
                else:
                    template = template[0]
            return template.format(*input_column_descriptions)

        # generic case:
        name = self.name.upper() if self.name is not None else type(self).__name__
        if slice_num is not None:
            nth_slice = convert_to_nth(slice_num + 1)
            description = "the {} output from applying {} to {}".format(
                nth_slice,
                name,
                ", ".join(input_column_descriptions),
            )
        else:
            description = "the result of applying {} to {}".format(
                name,
                ", ".join(input_column_descriptions),
            )
        return description

    @staticmethod
    def flatten_nested_input_types(input_types):
        """Flattens nested column schema inputs into a single list."""
        if isinstance(input_types[0], list):
            input_types = [
                sub_input for input_obj in input_types for sub_input in input_obj
            ]
        return input_types



import math
import re

import pandas as pd
from tqdm import tqdm
from woodwork.logical_types import Boolean, Categorical, Ordinal

import featuretools as ft


def load_flight(
    month_filter=None,
    categorical_filter=None,
    nrows=None,
    demo=True,
    return_single_table=False,
    verbose=False,
):
    """
    Download, clean, and filter flight data from 2017.
    The original dataset can be found `here <https://www.transtats.bts.gov/ot_delay/ot_delaycause1.asp>`_.

    Args:

        month_filter (list[int]): Only use data from these months (example is ``[1, 2]``).
            To skip, set to None.
        categorical_filter (dict[str->str]): Use only specified categorical values.
            Example is ``{'dest_city': ['Boston, MA'], 'origin_city': ['Boston, MA']}``
            which returns all flights in OR out of Boston. To skip, set to None.
        nrows (int): Passed to nrows in ``pd.read_csv``. Used before filtering.
        demo (bool): Use only two months of data. If False, use the whole year.
        return_single_table (bool): Exit the function early and return a dataframe.
        verbose (bool): Show a progress bar while loading the data.

    Examples:

        .. ipython::
            :verbatim:

            In [1]: import featuretools as ft

            In [2]: es = ft.demo.load_flight(verbose=True,
               ...:                          month_filter=[1],
               ...:                          categorical_filter={'origin_city':['Boston, MA']})
            100%|xxxxxxxxxxxxxxxxxxxxxxxxx| 100/100 [01:16<00:00,  1.31it/s]

            In [3]: es
            Out[3]:
            Entityset: Flight Data
              DataFrames:
                airports [Rows: 55, Columns: 3]
                flights [Rows: 613, Columns: 9]
                trip_logs [Rows: 9456, Columns: 22]
                airlines [Rows: 10, Columns: 1]
              Relationships:
                trip_logs.flight_id -> flights.flight_id
                flights.carrier -> airlines.carrier
                flights.dest -> airports.dest
    """

    filename, csv_length = get_flight_filename(demo=demo)

    print("Downloading data ...")
    url = "https://oss.alteryx.com/datasets/{}?library=featuretools&version={}".format(
        filename,
        ft.__version__,
    )

    chunksize = math.ceil(csv_length / 99)
    pd.options.display.max_columns = 200
    iter_csv = pd.read_csv(
        url,
        compression="zip",
        iterator=True,
        nrows=nrows,
        chunksize=chunksize,
    )
    if verbose:
        iter_csv = tqdm(iter_csv, total=100)

    partial_df_list = []
    for chunk in iter_csv:
        df = filter_data(
            _clean_data(chunk),
            month_filter=month_filter,
            categorical_filter=categorical_filter,
        )
        partial_df_list.append(df)
    data = pd.concat(partial_df_list)

    if return_single_table:
        return data

    es = make_es(data)

    return es


def make_es(data):
    es = ft.EntitySet("Flight Data")
    arr_time_columns = [
        "arr_delay",
        "dep_delay",
        "carrier_delay",
        "weather_delay",
        "national_airspace_delay",
        "security_delay",
        "late_aircraft_delay",
        "canceled",
        "diverted",
        "taxi_in",
        "taxi_out",
        "air_time",
        "dep_time",
    ]

    logical_types = {
        "flight_num": Categorical,
        "distance_group": Ordinal(order=[1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11]),
        "canceled": Boolean,
        "diverted": Boolean,
    }

    es.add_dataframe(
        data,
        dataframe_name="trip_logs",
        index="trip_log_id",
        make_index=True,
        time_index="date_scheduled",
        secondary_time_index={"arr_time": arr_time_columns},
        logical_types=logical_types,
    )

    es.normalize_dataframe(
        "trip_logs",
        "flights",
        "flight_id",
        additional_columns=[
            "origin",
            "origin_city",
            "origin_state",
            "dest",
            "dest_city",
            "dest_state",
            "distance_group",
            "carrier",
            "flight_num",
        ],
    )

    es.normalize_dataframe("flights", "airlines", "carrier", make_time_index=False)

    es.normalize_dataframe(
        "flights",
        "airports",
        "dest",
        additional_columns=["dest_city", "dest_state"],
        make_time_index=False,
    )
    return es


def _clean_data(data):
    # Make column names snake case
    clean_data = data.rename(columns={col: convert(col) for col in data})

    # Chance crs -> "scheduled" and other minor clarifications
    clean_data = clean_data.rename(
        columns={
            "crs_arr_time": "scheduled_arr_time",
            "crs_dep_time": "scheduled_dep_time",
            "crs_elapsed_time": "scheduled_elapsed_time",
            "nas_delay": "national_airspace_delay",
            "origin_city_name": "origin_city",
            "dest_city_name": "dest_city",
            "cancelled": "canceled",
        },
    )

    # Combine strings like 0130 (1:30 AM) with dates (2017-01-01)
    clean_data["scheduled_dep_time"] = clean_data["scheduled_dep_time"].apply(
        lambda x: str(x),
    ) + clean_data["flight_date"].astype("str")

    # Parse combined string as a date
    clean_data.loc[:, "scheduled_dep_time"] = pd.to_datetime(
        clean_data["scheduled_dep_time"],
        format="%H%M%Y-%m-%d",
        errors="coerce",
    )

    clean_data["scheduled_elapsed_time"] = pd.to_timedelta(
        clean_data["scheduled_elapsed_time"],
        unit="m",
    )

    clean_data = _reconstruct_times(clean_data)

    # Create a time index 6 months before scheduled_dep
    clean_data.loc[:, "date_scheduled"] = pd.to_datetime(
        clean_data["scheduled_dep_time"],
    ).dt.date - pd.Timedelta("120d")

    # A null entry for a delay means no delay
    clean_data = _fill_labels(clean_data)

    # Nulls for scheduled values are too problematic. Remove them.
    clean_data = clean_data.dropna(
        axis="rows",
        subset=["scheduled_dep_time", "scheduled_arr_time"],
    )

    # Make a flight id. Define a flight as a combination of:
    # 1. carrier 2. flight number 3. origin airport 4. dest airport
    clean_data.loc[:, "flight_id"] = (
        clean_data["carrier"]
        + "-"
        + clean_data["flight_num"].apply(lambda x: str(x))
        + ":"
        + clean_data["origin"]
        + "->"
        + clean_data["dest"]
    )

    column_order = [
        "flight_id",
        "flight_num",
        "date_scheduled",
        "scheduled_dep_time",
        "scheduled_arr_time",
        "carrier",
        "origin",
        "origin_city",
        "origin_state",
        "dest",
        "dest_city",
        "dest_state",
        "distance_group",
        "dep_time",
        "arr_time",
        "dep_delay",
        "taxi_out",
        "taxi_in",
        "arr_delay",
        "diverted",
        "scheduled_elapsed_time",
        "air_time",
        "distance",
        "carrier_delay",
        "weather_delay",
        "national_airspace_delay",
        "security_delay",
        "late_aircraft_delay",
        "canceled",
    ]

    clean_data = clean_data[column_order]

    return clean_data


def _fill_labels(clean_data):
    labely_columns = [
        "arr_delay",
        "dep_delay",
        "carrier_delay",
        "weather_delay",
        "national_airspace_delay",
        "security_delay",
        "late_aircraft_delay",
        "canceled",
        "diverted",
        "taxi_in",
        "taxi_out",
        "air_time",
    ]
    for col in labely_columns:
        clean_data.loc[:, col] = clean_data[col].fillna(0)

    return clean_data


def _reconstruct_times(clean_data):
    """Reconstruct departure_time, scheduled_dep_time,
    arrival_time and scheduled_arr_time by adding known delays
    to known times. We do:
        - dep_time is scheduled_dep + dep_delay
        - arr_time is dep_time + taxiing and air_time
        - scheduled arrival is scheduled_dep + scheduled_elapsed
    """
    clean_data.loc[:, "dep_time"] = clean_data["scheduled_dep_time"] + pd.to_timedelta(
        clean_data["dep_delay"],
        unit="m",
    )

    clean_data.loc[:, "arr_time"] = clean_data["dep_time"] + pd.to_timedelta(
        clean_data["taxi_out"] + clean_data["air_time"] + clean_data["taxi_in"],
        unit="m",
    )

    clean_data.loc[:, "scheduled_arr_time"] = (
        clean_data["scheduled_dep_time"] + clean_data["scheduled_elapsed_time"]
    )
    return clean_data


def filter_data(clean_data, month_filter=None, categorical_filter=None):
    if month_filter is not None:
        tmp = pd.to_datetime(clean_data["scheduled_dep_time"]).dt.month.isin(
            month_filter,
        )
        clean_data = clean_data[tmp]

    if categorical_filter is not None:
        tmp = False
        for key, values in categorical_filter.items():
            tmp = tmp | clean_data[key].isin(values)
        clean_data = clean_data[tmp]

    return clean_data


def convert(name):
    # Rename columns to underscore
    # Code via SO https://stackoverflow.com/questions/1175208/elegant-python-function-to-convert-camelcase-to-snake-case
    s1 = re.sub("(.)([A-Z][a-z]+)", r"\1_\2", name)
    return re.sub("([a-z0-9])([A-Z])", r"\1_\2", s1).lower()


def get_flight_filename(demo=True):
    if demo:
        filename = SMALL_FLIGHT_CSV
        rows = 860457
    else:
        filename = BIG_FLIGHT_CSV
        rows = 5162742

    return filename, rows


SMALL_FLIGHT_CSV = "data_2017_jan_feb.csv.zip"
BIG_FLIGHT_CSV = "data_all_2017.csv.zip"



from woodwork.column_schema import ColumnSchema
from woodwork.logical_types import BooleanNullable, Datetime

from featuretools.primitives.base import TransformPrimitive
from featuretools.utils.gen_utils import Library


class IsWorkingHours(TransformPrimitive):
    """Determines if a datetime falls during working hours on a 24-hour clock. Can configure start_hour and end_hour.

    Args:
        start_hour (int): Start hour of workday. Must adhere to 24-hour clock. Default is 8 (8am).
        end_hour (int): End hour of workday. Must adhere to 24-hour clock. Default is 18 (6pm).

    Examples:
        >>> import numpy as np
        >>> from datetime import datetime
        >>> dates = [datetime(2022, 6, 21, 16, 3, 3),
        ...          datetime(2019, 1, 3, 4, 4, 4),
        ...          datetime(2022, 1, 1, 12, 1, 2),
        ...          np.nan]
        >>> is_working_hour = IsWorkingHours()
        >>> is_working_hour(dates).tolist()
        [True, False, True, False]
        >>> is_working_hour = IsWorkingHours(15, 17)
        >>> is_working_hour(dates).tolist()
        [True, False, False, False]
    """

    name = "is_working_hours"
    input_types = [ColumnSchema(logical_type=Datetime)]
    return_type = ColumnSchema(logical_type=BooleanNullable)
    compatibility = [Library.PANDAS, Library.DASK, Library.SPARK]
    description_template = "whether {} falls during working hours"

    def __init__(self, start_hour=8, end_hour=18):
        self.start_hour = start_hour
        self.end_hour = end_hour

    def get_function(self):
        def is_working_hours(vals):
            return (vals.dt.hour >= self.start_hour) & (vals.dt.hour <= self.end_hour)

        return is_working_hours



import datetime
import json
import os
import tarfile
import tempfile

from woodwork.serializers.serializer_base import typing_info_to_dict

from featuretools.utils.gen_utils import import_or_none
from featuretools.utils.s3_utils import get_transport_params, use_smartopen_es
from featuretools.utils.wrangle import _is_s3, _is_url
from featuretools.version import ENTITYSET_SCHEMA_VERSION

ps = import_or_none("pyspark.pandas")

FORMATS = ["csv", "pickle", "parquet"]


def entityset_to_description(entityset, format=None):
    """Serialize entityset to data description.

    Args:
        entityset (EntitySet) : Instance of :class:`.EntitySet`.

    Returns:
        description (dict) : Description of :class:`.EntitySet`.
    """

    dataframes = {
        dataframe.ww.name: typing_info_to_dict(dataframe)
        for dataframe in entityset.dataframes
    }
    relationships = [
        relationship.to_dictionary() for relationship in entityset.relationships
    ]

    data_type = entityset.dataframe_type

    data_description = {
        "schema_version": ENTITYSET_SCHEMA_VERSION,
        "id": entityset.id,
        "dataframes": dataframes,
        "relationships": relationships,
        "format": format,
        "data_type": data_type,
    }
    return data_description


def write_data_description(entityset, path, profile_name=None, **kwargs):
    """Serialize entityset to data description and write to disk or S3 path.

    Args:
        entityset (EntitySet) : Instance of :class:`.EntitySet`.
        path (str) : Location on disk or S3 path to write `data_description.json` and dataframe data.
        profile_name (str, bool): The AWS profile specified to write to S3. Will default to None and search for AWS credentials.
            Set to False to use an anonymous profile.
        kwargs (keywords) : Additional keyword arguments to pass as keywords arguments to the underlying serialization method or to specify AWS profile.
    """
    if _is_s3(path):
        with tempfile.TemporaryDirectory() as tmpdir:
            os.makedirs(os.path.join(tmpdir, "data"))
            dump_data_description(entityset, tmpdir, **kwargs)
            file_path = create_archive(tmpdir)

            transport_params = get_transport_params(profile_name)
            use_smartopen_es(
                file_path,
                path,
                read=False,
                transport_params=transport_params,
            )
    elif _is_url(path):
        raise ValueError("Writing to URLs is not supported")
    else:
        path = os.path.abspath(path)
        os.makedirs(os.path.join(path, "data"), exist_ok=True)
        dump_data_description(entityset, path, **kwargs)


def dump_data_description(entityset, path, **kwargs):
    format = kwargs.get("format")
    description = entityset_to_description(entityset, format)
    for df in entityset.dataframes:
        data_path = os.path.join(path, "data", df.ww.name)
        os.makedirs(os.path.join(data_path, "data"), exist_ok=True)
        df.ww.to_disk(data_path, **kwargs)
    file = os.path.join(path, "data_description.json")
    with open(file, "w") as file:
        json.dump(description, file)


def create_archive(tmpdir):
    file_name = "es-{date:%Y-%m-%d_%H%M%S}.tar".format(date=datetime.datetime.now())
    file_path = os.path.join(tmpdir, file_name)
    tar = tarfile.open(str(file_path), "w")
    tar.add(str(tmpdir) + "/data_description.json", arcname="/data_description.json")
    tar.add(str(tmpdir) + "/data", arcname="/data")
    tar.close()
    return file_path



"""
cache.py

Custom caching class, currently used for FeatureBase
"""
# needed for defaultdict annotation if < python 3.9
from __future__ import annotations

from collections import defaultdict
from dataclasses import dataclass, field
from enum import Enum
from typing import Any, List, Optional, Union


class CacheType(Enum):
    """Enumerates the supported cache types"""

    DEPENDENCY = 1
    DEPTH = 2


@dataclass()
class FeatureCache:
    """Provides caching for the defined types"""

    enabled: bool = False
    cache: defaultdict[dict] = field(default_factory=lambda: defaultdict(dict))

    def get(
        self,
        cache_type: CacheType,
        hashkey: int,
    ) -> Optional[Union[List[Any], Any]]:
        """Gets the cache entry, if enabled and defined

        Args:
            cache_type (CacheType): type of cache
            hashkey (int): hash key

        Returns:
            Optional[Union[List[Any], Any]]: payload assigned to the hashkey
        """
        if not self.enabled or cache_type not in self.cache:
            return None
        return self.cache[cache_type].get(hashkey, None)

    def add(self, cache_type: CacheType, hashkey: int, payload: Any):
        """Adds an entry to the cache, if enabled

        Args:
            cache_type (CacheType): type of cache
            hashkey (int): hash key
            payload (Any): payload to assign
        """
        if self.enabled:
            self.cache[cache_type][hashkey] = payload

    def clear_all(self):
        """Clears the cache collections"""
        self.cache.clear()


feature_cache = FeatureCache()



from woodwork.column_schema import ColumnSchema
from woodwork.logical_types import Datetime, Integer

from featuretools.primitives.base import AggregationPrimitive


class NUniqueDaysOfCalendarYear(AggregationPrimitive):
    """Determines the number of unique calendar days.

    Description:
        Given a list of datetimes, return the number of unique calendar
        days. The same date in two different years is counted as one. So
        Feb 21, 2017 is not unique from Feb 21, 2019.

    Examples:
        >>> from datetime import datetime
        >>> n_unique_days_of_calendar_year = NUniqueDaysOfCalendarYear()
        >>> times = [datetime(2019, 2, 1),
        ...          datetime(2019, 2, 1),
        ...          datetime(2018, 2, 1),
        ...          datetime(2019, 1, 1)]
        >>> n_unique_days_of_calendar_year(times)
        2
    """

    name = "n_unique_days_of_calendar_year"
    input_types = [ColumnSchema(logical_type=Datetime)]
    return_type = ColumnSchema(logical_type=Integer, semantic_tags={"numeric"})
    stack_on_self = False
    default_value = 0

    def get_function(self):
        def n_unique_days_of_calendar_year(x):
            return x.dropna().dt.strftime("%m-%d").nunique()

        return n_unique_days_of_calendar_year



from math import floor

import numpy as np
from scipy.signal import savgol_coeffs, savgol_filter
from woodwork.column_schema import ColumnSchema
from woodwork.logical_types import Double

from featuretools.primitives.base import TransformPrimitive


class SavgolFilter(TransformPrimitive):
    """Applies a Savitzky-Golay filter to a list of values.

    Description:
        Given a list of values, return a smoothed list which increases
        the signal to noise ratio without greatly distoring the
        signal. Uses the `Savitzky–Golay filter` method.

        If the input list has less than 20 values, it will be returned
        as is.

        See the following page for more info:
        https://docs.scipy.org/doc/scipy-0.16.0/reference/generated/scipy.signal.savgol_filter.html

    Args:
        window_length (int):  The length of the filter window (i.e. the number
            of coefficients). `window_length` must be a positive odd integer.

        polyorder (int): The order of the polynomial used to fit the samples.
            `polyorder` must be less than `window_length`.

        deriv (int): Optional. The order of the derivative to compute.  This
            must be a nonnegative integer.  The default is 0, which means to
            filter the data without differentiating.

        delta (float): Optional. The spacing of the samples to which the filter
            will be applied. This is only used if deriv > 0.  Default is 1.0.

        mode (str): Optional. Must be 'mirror', 'constant', 'nearest', 'wrap'
            or 'interp'.  This determines the type of extension to use for the
            padded signal to which the filter is applied.  When `mode` is
            'constant', the padding value is given by `cval`.  See the Notes
            for more details on 'mirror', 'constant', 'wrap', and 'nearest'.

            When the 'interp' mode is selected (the default), no extension
            is used.  Instead, a degree `polyorder` polynomial is fit to the
            last `window_length` values of the edges, and this polynomial is
            used to evaluate the last `window_length // 2` output values.

        cval (scalar): Optional. Value to fill past the edges of the input
            if `mode` is 'constant'. Default is 0.0.

    Examples:
        >>> savgol_filter = SavgolFilter()
        >>> data = [0, 1, 1, 2, 3, 4, 5, 7, 8, 7, 9, 9, 12, 11, 12, 14, 15, 17, 17, 17, 20]
        >>> [round(x, 4) for x in savgol_filter(data).tolist()[:3]]
        [0.0429, 0.8286, 1.2571]

        We can control `window_length` and `polyorder` of the filter.

        >>> savgol_filter = SavgolFilter(window_length=13, polyorder=3)
        >>> [round(x, 4) for x in savgol_filter(data).tolist()[:3]]
        [-0.0962, 0.6484, 1.4451]

        We can also control the `deriv` and `delta` parameters.

        >>> savgol_filter = SavgolFilter(deriv=1, delta=1.5)
        >>> [round(x, 4) for x in savgol_filter(data).tolist()[:3]]
        [0.754, 0.3492, 0.2778]

        Finally, we can use `mode` to control how edge values are handled.

        >>> savgol_filter = SavgolFilter(mode='constant', cval=5)
        >>> [round(x, 4) for x in savgol_filter(data).tolist()[:3]]
        [1.5429, 0.2286, 1.2571]
    """

    name = "savgol_filter"
    input_types = [ColumnSchema(semantic_tags={"numeric"})]
    return_type = ColumnSchema(logical_type=Double, semantic_tags={"numeric"})

    def __init__(
        self,
        window_length=None,
        polyorder=None,
        deriv=0,
        delta=1.0,
        mode="interp",
        cval=0.0,
    ):
        if window_length is not None and polyorder is not None:
            try:
                if mode not in ["mirror", "constant", "nearest", "interp", "wrap"]:
                    raise ValueError(
                        "mode must be 'mirror', 'constant', "
                        "'nearest', 'wrap' or 'interp'.",
                    )
                savgol_coeffs(window_length, polyorder, deriv=deriv, delta=delta)
            except Exception:
                raise
        elif (window_length is None and polyorder is not None) or (
            window_length is not None and polyorder is None
        ):
            error_text = (
                "Both window_length and polyorder must be defined if you define one."
            )
            raise ValueError(error_text)

        self.window_length = window_length
        self.polyorder = polyorder
        self.deriv = deriv
        self.delta = delta
        self.mode = mode
        self.cval = cval

    def get_function(self):
        def smooth(x):
            if x.shape[0] < 20:
                return x
            if np.isnan(np.min(x)):
                # interpolate the nan values, works for edges & middle nans
                mask = np.isnan(x)
                x[mask] = np.interp(
                    np.flatnonzero(mask),
                    np.flatnonzero(~mask),
                    x[~mask],
                )
            window_length = self.window_length
            polyorder = self.polyorder
            if window_length is None and polyorder is None:
                window_length = floor(len(x) / 10) * 2 + 1
                polyorder = 3
            return savgol_filter(
                x,
                window_length=window_length,
                polyorder=polyorder,
                deriv=self.deriv,
                delta=self.delta,
                mode=self.mode,
                cval=self.cval,
            )

        return smooth
